import pytest
import requests
import responses
from responses import matchers

from pbipy.dataflows import Dataflow


@pytest.fixture
def dataflow():
    raw = {
        "objectId": "928228ba-008d-4fd9-864a-92d2752ee5ce",
        "name": "AdventureWorks",
        "description": "Our Adventure Works",
        "modelUrl": "https://MyDataflowStorageAccount.dfs.core.windows.net/powerbi/contoso/AdventureWorks/model.json",
    }

    dataflow = Dataflow(
        id=raw.get("objectId"),
        session=requests.Session(),
        group_id="51e47fc5-48fd-4826-89f0-021bd3a80abd",
        raw=raw,
    )

    return dataflow


def test_dataflow_creation(dataflow):
    assert isinstance(dataflow, Dataflow)
    assert dataflow.id == "928228ba-008d-4fd9-864a-92d2752ee5ce"
    assert dataflow.name == "AdventureWorks"
    assert dataflow.description == "Our Adventure Works"


@responses.activate
def test_datasources(dataflow, get_dataflow_datasources):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups/51e47fc5-48fd-4826-89f0-021bd3a80abd/dataflows/928228ba-008d-4fd9-864a-92d2752ee5ce/datasources",
        body=get_dataflow_datasources,
        content_type="application/json",
    )

    datasources = dataflow.datasources()

    assert isinstance(datasources, list)
    assert all(isinstance(datasource, dict) for datasource in datasources)


@responses.activate
def test_transactions(dataflow, get_dataflow_transactions):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups/51e47fc5-48fd-4826-89f0-021bd3a80abd/dataflows/928228ba-008d-4fd9-864a-92d2752ee5ce/transactions",
        body=get_dataflow_transactions,
        content_type="application/json",
    )

    transactions = dataflow.transactions()

    assert isinstance(transactions, list)
    assert all(isinstance(transaction, dict) for transaction in transactions)
    assert len(transactions) == 2


# TODO: Add test that includes response body
@responses.activate
def test_upstream_dataflows(dataflow):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups/51e47fc5-48fd-4826-89f0-021bd3a80abd/dataflows/928228ba-008d-4fd9-864a-92d2752ee5ce/upstreamDataflows",
        body='{"value":[]}',
    )

    dataflow.upstream_dataflows()


@responses.activate
def test_refresh_no_params(dataflow):
    json_params = {"notifyOption": "MailOnCompletion"}

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/groups/51e47fc5-48fd-4826-89f0-021bd3a80abd/dataflows/928228ba-008d-4fd9-864a-92d2752ee5ce/refreshes",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    dataflow.refresh(notify_option="MailOnCompletion")


@responses.activate
def test_refresh_params(dataflow):
    json_params = {"notifyOption": "MailOnCompletion"}

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/groups/51e47fc5-48fd-4826-89f0-021bd3a80abd/dataflows/928228ba-008d-4fd9-864a-92d2752ee5ce/refreshes?processType=default",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    dataflow.refresh(notify_option="MailOnCompletion", process_type="default")


@responses.activate
def test_update_call(dataflow):
    json_params = {
        "name": "New Dataflow Name",
        "allowNativeQueries": True,
    }

    responses.patch(
        "https://api.powerbi.com/v1.0/myorg/groups/51e47fc5-48fd-4826-89f0-021bd3a80abd/dataflows/928228ba-008d-4fd9-864a-92d2752ee5ce",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    dataflow.update(
        name="New Dataflow Name",
        allow_native_queries=True,
    )


def test_update_raises(dataflow):
    with pytest.raises(ValueError):
        dataflow.update()


@responses.activate
def test_update_refresh_schedule(dataflow):
    json_params = {
        "value": {
            "days": [
                "Monday",
                "Wednesday",
            ],
            "times": [
                "10:00",
                "16:00",
            ],
            "notifyOption": "NoNotification",
        }
    }

    responses.patch(
        "https://api.powerbi.com/v1.0/myorg/groups/51e47fc5-48fd-4826-89f0-021bd3a80abd/dataflows/928228ba-008d-4fd9-864a-92d2752ee5ce/refreshSchedule",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    dataflow.update_refresh_schedule(
        days=["Monday", "Wednesday"],
        times=["10:00", "16:00"],
        notify_option="NoNotification",
    )


def test_update_refresh_schedule_raises(dataflow):
    with pytest.raises(ValueError):
        dataflow.update_refresh_schedule()



from io import BytesIO
import pathlib
from unittest.mock import mock_open, patch

import pytest
import requests
import responses
from responses import matchers
from requests.exceptions import HTTPError

from pbipy.apps import App
from pbipy.dashboards import Dashboard
from pbipy.dataflows import Dataflow
from pbipy.groups import Group
from pbipy.imports import Import, TemporaryUploadLocation
from pbipy.reports import Report


@responses.activate
def test_get_dataset_calls_correct_get_dataset_url(powerbi, get_dataset):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        body=get_dataset,
        content_type="application/json",
    )

    powerbi.dataset("cfafbeb1-8037-4d0c-896e-a46fb27ff229")


@responses.activate
def test_get_dataset_calls_correct_get_dataset_in_group_url(powerbi, get_dataset):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        body=get_dataset,
        content_type="application/json",
    )

    powerbi.dataset(
        "cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        group="f089354e-8366-4e18-aea3-4cb4a3a50b48",
    )


@responses.activate
def test_get_dataset_properties_are_set(powerbi, get_dataset):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        body=get_dataset,
        content_type="application/json",
    )

    dataset = powerbi.dataset("cfafbeb1-8037-4d0c-896e-a46fb27ff229")

    assert dataset.id == "cfafbeb1-8037-4d0c-896e-a46fb27ff229"
    assert dataset.name == "SalesMarketing"
    assert dataset.add_rows_api_enabled == False
    assert dataset.configured_by == "john@contoso.com"
    assert dataset.group_id == None
    assert dataset.is_refreshable == True


@responses.activate
def test_get_dataset_group_property_is_set(powerbi, get_dataset):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        body=get_dataset,
        content_type="application/json",
    )

    dataset = powerbi.dataset(
        "cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        group="f089354e-8366-4e18-aea3-4cb4a3a50b48",
    )

    assert dataset.group_id == "f089354e-8366-4e18-aea3-4cb4a3a50b48"


@responses.activate
def test_get_datasets_calls_correct_get_datasets_url(powerbi, get_datasets):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/datasets",
        body=get_datasets,
        content_type="application/json",
    )

    powerbi.datasets()


@responses.activate
def test_get_datasets_calls_correct_get_datasets_in_group_url(powerbi, get_datasets):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/datasets",
        body=get_datasets,
        content_type="application/json",
    )

    powerbi.datasets(group="f089354e-8366-4e18-aea3-4cb4a3a50b48")


@responses.activate
def test_get_datasets_sets_group_property(powerbi, get_datasets):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/datasets",
        body=get_datasets,
        content_type="application/json",
    )

    datasets = powerbi.datasets(group="f089354e-8366-4e18-aea3-4cb4a3a50b48")

    assert all(hasattr(dataset, "group_id") for dataset in datasets)
    assert all(
        getattr(dataset, "group_id") == "f089354e-8366-4e18-aea3-4cb4a3a50b48"
        for dataset in datasets
    )


@responses.activate
def test_get_datasets_sets_js_property(powerbi, get_datasets):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/datasets",
        body=get_datasets,
        content_type="application/json",
    )

    datasets = powerbi.datasets(group="f089354e-8366-4e18-aea3-4cb4a3a50b48")

    assert all(hasattr(dataset, "raw") for dataset in datasets)


@responses.activate
def test_delete_dataset_call(powerbi):
    responses.delete(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229"
    )

    powerbi.delete_dataset(
        "cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        group="f089354e-8366-4e18-aea3-4cb4a3a50b48",
    )


@responses.activate
def test_delete_dataset_raises_http_error(powerbi):
    responses.delete(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        status=400,
        body="{}",
    )

    with pytest.raises(HTTPError):
        powerbi.delete_dataset(
            "cfafbeb1-8037-4d0c-896e-a46fb27ff229",
            group="f089354e-8366-4e18-aea3-4cb4a3a50b48",
        )


@responses.activate
def test_groups_call_with_params(powerbi, get_groups):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups?$filter=contains(name,'marketing')%20or%20name%20eq%20'contoso'",
        body=get_groups,
        content_type="application/json",
    )

    powerbi.groups(filter="contains(name,'marketing') or name eq 'contoso'")


@responses.activate
def test_groups_result(powerbi, get_groups):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups",
        body=get_groups,
        content_type="application/json",
    )

    groups = powerbi.groups()

    assert isinstance(groups, list)
    assert all(isinstance(group, Group) for group in groups)


@responses.activate
def test_group(powerbi, get_group):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups?$filter=id%20eq%20'a2f89923-421a-464e-bf4c-25eab39bb09f'",
        body=get_group,
        content_type="application/json",
    )

    powerbi.group("a2f89923-421a-464e-bf4c-25eab39bb09f")


@responses.activate
def test_group_raises(powerbi):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups?$filter=id%20eq%20'a2f89923-421a-464e-bf4c-25eab39bb09f'",
        body="{value:[]}",
        content_type="application/json",
    )

    with pytest.raises(ValueError):
        powerbi.group("a2f89923-421a-464e-bf4c-25eab39bb09f")


@responses.activate
def test_delete_group_call_with_group_object(powerbi):
    responses.delete(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48"
    )

    group = Group(
        "f089354e-8366-4e18-aea3-4cb4a3a50b48",
        session=requests.Session(),
    )

    powerbi.delete_group(group)


@responses.activate
def test_delete_group_call(powerbi):
    responses.delete(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48"
    )

    powerbi.delete_group("f089354e-8366-4e18-aea3-4cb4a3a50b48")


@responses.activate
def test_report_call(powerbi, get_report):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/reports/5b218778-e7a5-4d73-8187-f10824047715",
        body=get_report,
        content_type="application/json",
    )

    powerbi.report("5b218778-e7a5-4d73-8187-f10824047715")


@responses.activate
def test_report_call_with_group(powerbi, get_report):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/reports/5b218778-e7a5-4d73-8187-f10824047715",
        body=get_report,
        content_type="application/json",
    )

    powerbi.report(
        report="5b218778-e7a5-4d73-8187-f10824047715",
        group="f089354e-8366-4e18-aea3-4cb4a3a50b48",
    )


@responses.activate
def test_report_call_result(powerbi, get_report):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/reports/5b218778-e7a5-4d73-8187-f10824047715",
        body=get_report,
        content_type="application/json",
    )

    report = powerbi.report(
        report="5b218778-e7a5-4d73-8187-f10824047715",
        group="f089354e-8366-4e18-aea3-4cb4a3a50b48",
    )

    assert isinstance(report, Report)
    assert hasattr(report, "group_id")
    assert report.group_id == "f089354e-8366-4e18-aea3-4cb4a3a50b48"
    assert report.name == "SalesMarketing"


@responses.activate
def test_reports_call(powerbi, get_reports_in_group):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/reports",
        body=get_reports_in_group,
        content_type="application/json",
    )

    powerbi.reports(group="f089354e-8366-4e18-aea3-4cb4a3a50b48")


@responses.activate
def test_reports_call_group_object(powerbi, get_reports_in_group):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/reports",
        body=get_reports_in_group,
        content_type="application/json",
    )

    group = Group(
        "f089354e-8366-4e18-aea3-4cb4a3a50b48",
        requests.Session(),
    )

    powerbi.reports(group=group)


@responses.activate
def test_reports_result(powerbi, get_reports_in_group):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/reports",
        body=get_reports_in_group,
        content_type="application/json",
    )

    reports = powerbi.reports(group="f089354e-8366-4e18-aea3-4cb4a3a50b48")

    assert isinstance(reports, list)
    assert all(isinstance(report, Report) for report in reports)


@responses.activate
def test_delete_report_call_report_str_no_group(powerbi):
    responses.delete(
        "https://api.powerbi.com/v1.0/myorg/reports/5b218778-e7a5-4d73-8187-f10824047715",
    )

    powerbi.delete_report("5b218778-e7a5-4d73-8187-f10824047715")


@responses.activate
def test_delete_report_call_report_object_ignore_group(powerbi):
    responses.delete(
        "https://api.powerbi.com/v1.0/myorg/reports/5b218778-e7a5-4d73-8187-f10824047715",
    )

    report = Report(
        id="5b218778-e7a5-4d73-8187-f10824047715",
        session=requests.Session(),
    )

    powerbi.delete_report(
        report,
        group="f089354e-8366-4e18-aea3-4cb4a3a50b48",
    )


@responses.activate
def test_delete_report_call_report_str_group_object(powerbi):
    responses.delete(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/reports/5b218778-e7a5-4d73-8187-f10824047715",
    )

    group = Group(
        "f089354e-8366-4e18-aea3-4cb4a3a50b48",
        requests.Session(),
    )

    powerbi.delete_report(
        "5b218778-e7a5-4d73-8187-f10824047715",
        group=group,
    )


@responses.activate
def test_delete_report_call_report_object_with_group_ignore_group_id(powerbi):
    responses.delete(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/reports/5b218778-e7a5-4d73-8187-f10824047715",
    )

    report = Report(
        id="5b218778-e7a5-4d73-8187-f10824047715",
        group_id="f089354e-8366-4e18-aea3-4cb4a3a50b48",
        session=requests.Session(),
    )

    powerbi.delete_report(
        report,
        group="3d9b93c6-7b6d-4801-a491-1738910904fd",
    )


@responses.activate
def test_app(get_app, powerbi):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/apps/f089354e-8366-4e18-aea3-4cb4a3a50b48",
        body=get_app,
        content_type="application/json",
    )

    app = powerbi.app("f089354e-8366-4e18-aea3-4cb4a3a50b48")

    assert isinstance(app, App)
    assert app.id == "f089354e-8366-4e18-aea3-4cb4a3a50b48"


@responses.activate
def test_apps(get_apps, powerbi):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/apps",
        body=get_apps,
        content_type="application/json",
    )

    apps = powerbi.apps()

    assert isinstance(apps, list)
    assert all(isinstance(app, App) for app in apps)
    assert len(apps) == 2


@responses.activate
def test_dataflow(powerbi, get_dataflow):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/dataflows/bd32e5c0-363f-430b-a03b-5535a4804b9b",
        body=get_dataflow,
        content_type="application/json",
    )

    dataflow = powerbi.dataflow(
        "bd32e5c0-363f-430b-a03b-5535a4804b9b",
        group="f089354e-8366-4e18-aea3-4cb4a3a50b48",
    )

    assert isinstance(dataflow, Dataflow)
    assert dataflow.id == "bd32e5c0-363f-430b-a03b-5535a4804b9b"
    assert dataflow.group_id == "f089354e-8366-4e18-aea3-4cb4a3a50b48"
    assert hasattr(dataflow, "ppdf_output_file_format")


@responses.activate
def test_dataflow_no_id(powerbi, get_dataflow_no_id):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/dataflows/bd32e5c0-363f-430b-a03b-5535a4804b9b",
        body=get_dataflow_no_id,
        content_type="application/json",
    )

    dataflow = powerbi.dataflow(
        "bd32e5c0-363f-430b-a03b-5535a4804b9b",
        group="f089354e-8366-4e18-aea3-4cb4a3a50b48",
    )

    assert isinstance(dataflow, Dataflow)
    assert dataflow.id == "bd32e5c0-363f-430b-a03b-5535a4804b9b"
    assert dataflow.group_id == "f089354e-8366-4e18-aea3-4cb4a3a50b48"
    assert hasattr(dataflow, "ppdf_output_file_format")


@responses.activate
def test_dataflows(powerbi, get_dataflows):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups/a2f89923-421a-464e-bf4c-25eab39bb09f/dataflows",
        body=get_dataflows,
        content_type="application/json",
    )

    dataflows = powerbi.dataflows("a2f89923-421a-464e-bf4c-25eab39bb09f")

    assert isinstance(dataflows, list)
    assert all(isinstance(dataflow, Dataflow) for dataflow in dataflows)


@responses.activate
def test_cancel_transaction(powerbi, cancel_dataflow_transaction):
    responses.post(
        "https://api.powerbi.com/v1.0/myorg/groups/51e47fc5-48fd-4826-89f0-021111110abd/dataflows/transactions/2020-09-11T19:21:52.8778432Z@9cc7a369-6112-4dba-97b6-b07ff5699568$1374282/cancel",
        body=cancel_dataflow_transaction,
        content_type="application/json",
    )

    cancelled_transaction = powerbi.cancel_transaction(
        transaction_id="2020-09-11T19:21:52.8778432Z@9cc7a369-6112-4dba-97b6-b07ff5699568$1374282",
        group="51e47fc5-48fd-4826-89f0-021111110abd",
    )

    assert isinstance(cancelled_transaction, dict)
    assert cancelled_transaction.get("status") == "SuccessfullyMarked"


@responses.activate
def test_delete_dataflow(powerbi):
    responses.delete(
        "https://api.powerbi.com/v1.0/myorg/groups/51e47fc5-48fd-4826-89f0-021bd3a80abd/dataflows/928228ba-008d-4fd9-864a-92d2752ee5ce"
    )

    powerbi.delete_dataflow(
        dataflow="928228ba-008d-4fd9-864a-92d2752ee5ce",
        group="51e47fc5-48fd-4826-89f0-021bd3a80abd",
    )


@responses.activate
def test_add_dashboard(powerbi, add_dashboard):
    json_params = {"name": "SalesMarketing"}

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/dashboards",
        body=add_dashboard,
        match=[
            matchers.json_params_matcher(json_params),
        ],
        content_type="application/json",
    )

    dashboard = powerbi.add_dashboard("SalesMarketing")

    assert isinstance(dashboard, Dashboard)
    assert dashboard.group_id is None
    assert dashboard.id == "69ffaa6c-b36d-4d01-96f5-1ed67c64d4af"


@responses.activate
def test_add_dashboard_with_group(powerbi, add_dashboard):
    json_params = {"name": "SalesMarketing"}

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/dashboards",
        body=add_dashboard,
        match=[
            matchers.json_params_matcher(json_params),
        ],
        content_type="application/json",
    )

    dashboard = powerbi.add_dashboard(
        "SalesMarketing",
        group="f089354e-8366-4e18-aea3-4cb4a3a50b48",
    )

    assert isinstance(dashboard, Dashboard)
    assert dashboard.group_id == "f089354e-8366-4e18-aea3-4cb4a3a50b48"
    assert dashboard.id == "69ffaa6c-b36d-4d01-96f5-1ed67c64d4af"


@responses.activate
def test_dashboard(powerbi, get_dashboard):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/dashboards/69ffaa6c-b36d-4d01-96f5-1ed67c64d4af",
        body=get_dashboard,
        content_type="application/json",
    )

    dashboard = powerbi.dashboard("69ffaa6c-b36d-4d01-96f5-1ed67c64d4af")

    assert isinstance(dashboard, Dashboard)
    assert dashboard.display_name == "SalesMarketing"
    assert dashboard.group_id is None


@responses.activate
def test_dashboard_with_group(powerbi, get_dashboard_in_group):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/dashboards/69ffaa6c-b36d-4d01-96f5-1ed67c64d4af",
        body=get_dashboard_in_group,
        content_type="application/json",
    )

    group = Group("f089354e-8366-4e18-aea3-4cb4a3a50b48", requests.Session())
    dashboard = powerbi.dashboard(
        "69ffaa6c-b36d-4d01-96f5-1ed67c64d4af",
        group,
    )

    assert isinstance(dashboard, Dashboard)
    assert dashboard.display_name == "SalesMarketing"
    assert dashboard.group_id == group.id


@responses.activate
def test_imported_file(powerbi, get_import):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/imports/82d9a37a-2b45-4221-b012-cb109b8e30c7",
        body=get_import,
        content_type="application/json",
    )

    imported_file = powerbi.imported_file("82d9a37a-2b45-4221-b012-cb109b8e30c7")

    assert isinstance(imported_file, Import)
    assert imported_file.import_state == "Succeeded"
    assert imported_file.name == "SalesMarketing"
    assert imported_file.connection_type == "import"
    assert isinstance(imported_file.datasets, list)


@responses.activate
def test_imported_file_with_group(powerbi, get_import):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/imports/82d9a37a-2b45-4221-b012-cb109b8e30c7",
        body=get_import,
        content_type="application/json",
    )

    imported_file = powerbi.imported_file(
        import_id="82d9a37a-2b45-4221-b012-cb109b8e30c7",
        group="f089354e-8366-4e18-aea3-4cb4a3a50b48",
    )

    assert imported_file.group_id == "f089354e-8366-4e18-aea3-4cb4a3a50b48"


@responses.activate
def test_imported_files(powerbi, get_imports):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/imports",
        body=get_imports,
        content_type="application/json",
    )

    imported_files = powerbi.imported_files()

    assert isinstance(imported_files, list)
    assert all(isinstance(imported_file, Import) for imported_file in imported_files)
    assert imported_files[0].id == "82d9a37a-2b45-4221-b012-cb109b8e30c7"
    assert imported_files[0].import_state == "Succeeded"
    assert imported_files[0].name == "SalesMarketing"


@responses.activate
def test_create_temporary_upload_location(powerbi, create_temporary_upload_location):
    responses.post(
        "https://api.powerbi.com/v1.0/myorg/imports/createTemporaryUploadLocation",
        body=create_temporary_upload_location,
    )

    temporary_upload_location = powerbi.create_temporary_upload_location()

    assert isinstance(temporary_upload_location, TemporaryUploadLocation)
    assert temporary_upload_location.url == "https://anotherexample.com"
    assert temporary_upload_location.expiration_time == "2024-01-01T12:00:00.1234567Z"


@responses.activate
def test_create_temporary_upload_location_with_group(
    powerbi, create_temporary_upload_location
):
    responses.post(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/imports/createTemporaryUploadLocation",
        body=create_temporary_upload_location,
    )

    temporary_upload_location = powerbi.create_temporary_upload_location(
        group="f089354e-8366-4e18-aea3-4cb4a3a50b48",
    )

    assert isinstance(temporary_upload_location, TemporaryUploadLocation)
    assert temporary_upload_location.url == "https://anotherexample.com"
    assert temporary_upload_location.expiration_time == "2024-01-01T12:00:00.1234567Z"


@responses.activate
def test_import_file_filepath(
    powerbi,
    post_import,
    get_import,
):
    file_contents = b"pbix contents"
    req_files = {"file": file_contents}
    params = {
        "datasetDisplayName": "file.pbix",
        "nameConflict": "Abort",
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/imports",
        body=post_import,
        match=[
            matchers.multipart_matcher(files=req_files),
            matchers.query_param_matcher(params=params),
        ],
    )

    responses.get(
        "https://api.powerbi.com/v1.0/myorg/imports/82d9a37a-2b45-4221-b012-cb109b8e30c7",
        body=get_import,
        content_type="application/json",
    )

    test_path = pathlib.Path("/a/nonexistent/file.pbix")

    with patch("builtins.open", mock_open(read_data=file_contents)) as mock_file:
        my_import = powerbi.import_file(
            test_path,
            dataset_display_name="file.pbix",
            name_conflict="Abort",
        )

    mock_file.assert_called_with(test_path, "rb")
    assert my_import.id == "82d9a37a-2b45-4221-b012-cb109b8e30c7"
    assert my_import.import_state == "Succeeded"


@responses.activate
def test_import_file_filelike(
    powerbi,
    post_import,
    get_import,
):
    file_contents = b"pbix contents"
    filelike = BytesIO(file_contents)
    req_files = {"file": file_contents}

    params = {
        "datasetDisplayName": "file.pbix",
        "nameConflict": "Abort",
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/imports",
        body=post_import,
        match=[
            matchers.multipart_matcher(files=req_files),
            matchers.query_param_matcher(params=params),
        ],
    )

    responses.get(
        "https://api.powerbi.com/v1.0/myorg/imports/82d9a37a-2b45-4221-b012-cb109b8e30c7",
        body=get_import,
        content_type="application/json",
    )

    my_import = powerbi.import_file(
        filelike,
        dataset_display_name="file.pbix",
        name_conflict="Abort",
    )

    assert my_import.id == "82d9a37a-2b45-4221-b012-cb109b8e30c7"
    assert my_import.import_state == "Succeeded"


@responses.activate
def test_import_large_file_filepath(
    powerbi,
    create_temporary_upload_location,
    post_import,
    get_import,
):
    # Create temporary upload location
    responses.post(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/imports/createTemporaryUploadLocation",
        body=create_temporary_upload_location,
    )

    # Upload the file to the shared access signature url
    file_contents = b"pbix contents"
    req_files = {"file": file_contents}
    responses.post(
        "https://anotherexample.com",
        match=[matchers.multipart_matcher(files=req_files)],
    )

    # Post the file details
    json_params = {"fileUrl": "https://anotherexample.com"}
    params = {
        "datasetDisplayName": "file.pbix",
        "nameConflict": "Abort",
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/imports",
        body=post_import,
        match=[
            matchers.json_params_matcher(params=json_params),
            matchers.query_param_matcher(params=params),
        ],
    )

    # Get the details
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/imports/82d9a37a-2b45-4221-b012-cb109b8e30c7",
        body=get_import,
        content_type="application/json",
    )

    with patch("builtins.open", mock_open(read_data=file_contents)) as mock_file:
        test_path = pathlib.Path("/a/nonexistent/file.pbix")

        my_import = powerbi.import_large_file(
            test_path,
            dataset_display_name="file.pbix",
            name_conflict="Abort",
            group="f089354e-8366-4e18-aea3-4cb4a3a50b48",
        )

    mock_file.assert_called_with(test_path, "rb")
    assert my_import.id == "82d9a37a-2b45-4221-b012-cb109b8e30c7"
    assert my_import.import_state == "Succeeded"
    assert my_import.group_id == "f089354e-8366-4e18-aea3-4cb4a3a50b48"


@responses.activate
def test_import_large_file_filelike(
    powerbi,
    create_temporary_upload_location,
    post_import,
    get_import,
):
    # Create temporary upload location
    responses.post(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/imports/createTemporaryUploadLocation",
        body=create_temporary_upload_location,
    )

    # Upload the file to the shared access signature url
    file_contents = b"pbix contents"
    filelike = BytesIO(file_contents)
    req_files = {"file": file_contents}

    responses.post(
        "https://anotherexample.com",
        match=[matchers.multipart_matcher(files=req_files)],
    )

    # Post the file details
    json_params = {"fileUrl": "https://anotherexample.com"}
    params = {
        "datasetDisplayName": "file.pbix",
        "nameConflict": "Abort",
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/imports",
        body=post_import,
        match=[
            matchers.json_params_matcher(params=json_params),
            matchers.query_param_matcher(params=params),
        ],
    )

    # Get the details
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/imports/82d9a37a-2b45-4221-b012-cb109b8e30c7",
        body=get_import,
        content_type="application/json",
    )

    my_import = powerbi.import_large_file(
        filelike,
        dataset_display_name="file.pbix",
        name_conflict="Abort",
        group="f089354e-8366-4e18-aea3-4cb4a3a50b48",
    )

    assert my_import.id == "82d9a37a-2b45-4221-b012-cb109b8e30c7"
    assert my_import.import_state == "Succeeded"
    assert my_import.group_id == "f089354e-8366-4e18-aea3-4cb4a3a50b48"



from pathlib import Path
import pytest
import requests
import responses
from requests.exceptions import HTTPError
from responses import matchers

from pbipy import _utils
from pbipy.groups import Group
from pbipy.reports import Report

# TODO: Tests for request_mixin.delete()
# TODO: Tests for request_mixin.get()


def test_to_snake_case():
    assert (_utils.to_snake_case("FluffyWuffy")) == "fluffy_wuffy"
    assert (_utils.to_snake_case("webURL")) == "web_url"
    assert _utils.to_snake_case("AyeBeeCee") == "aye_bee_cee"


def test_to_camel_case():
    assert (
        _utils.to_camel_case("dataset_user_access_right")
    ) == "datasetUserAccessRight"
    assert (_utils.to_camel_case("identifier")) == "identifier"
    assert (_utils.to_camel_case("principal_type")) == "principalType"


def test_remove_no_values():
    test_d = {
        "queries": [
            {
                "query": "EVALUATE VALUES(MyTable)",
            },
        ],
        "serializerSettings": {"includeNulls": None},
        "impersonatedUserName": "someuser@mycompany.com",
    }

    expected_d = {
        "queries": [
            {
                "query": "EVALUATE VALUES(MyTable)",
            },
        ],
        "impersonatedUserName": "someuser@mycompany.com",
    }

    assert _utils.remove_no_values(test_d) == expected_d


def test_file_path_from_components():
    expected_path = Path("sample_report.pbix")

    path = _utils.file_path_from_components(
        file_name="sample_report",
        extension="pbix",
    )

    assert path == expected_path
    assert isinstance(path, Path)


def test_file_path_from_components_path_object():
    expected_path = Path("C:/sample_dir/sample_report.pbix")

    directory = Path("C:/sample_dir")
    path = _utils.file_path_from_components(
        file_name="sample_report",
        directory=directory,
        extension="pbix",
    )

    assert path == expected_path
    assert isinstance(path, Path)


def test_file_path_from_components_str_directory():
    expected_path = Path("C:/sample_dir/sample_report.pbix")

    path = _utils.file_path_from_components(
        file_name="sample_report",
        directory="C:/sample_dir",
        extension="pbix",
    )

    assert path == expected_path
    assert isinstance(path, Path)


def test_file_path_from_components_dot_ext():
    expected_path = Path("C:/sample_dir/sample_report.pbix")

    path = _utils.file_path_from_components(
        file_name="sample_report",
        directory="C:/sample_dir",
        extension=".pbix",
    )

    assert path == expected_path
    assert isinstance(path, Path)


def test_to_identifier():
    assert _utils.to_identifier("$type") == "type"
    assert _utils.to_identifier("pbi:mashup") == "pbi_mashup"
    assert _utils.to_identifier("ppdf:outputFileFormat") == "ppdf_outputFileFormat"


@pytest.fixture
def session():
    return requests.Session()


# TODO: Add test for post_raw()


@responses.activate
def test_post(session):
    json_params = {
        "identifier": "john@contoso.com",
        "principalType": "User",
        "datasetUserAccessRight": "Read",
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/users",
        match=[matchers.json_params_matcher(json_params)],
    )

    resource = "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/users"
    payload = {
        "identifier": "john@contoso.com",
        "principalType": "User",
        "datasetUserAccessRight": "Read",
    }

    _utils.post(resource, session, payload)


@responses.activate
def test_post_raises(session):
    json_params = {
        "identifier": "john@contoso.com",
        "principalType": "User",
        "datasetUserAccessRight": "Read",
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/users",
        match=[matchers.json_params_matcher(json_params)],
        body="{}",
        status=400,
    )

    resource = "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/users"
    payload = {
        "identifier": "john@contoso.com",
        "principalType": "User",
        "datasetUserAccessRight": "Read",
    }

    with pytest.raises(HTTPError):
        _utils.post(resource, session, payload)


@pytest.fixture
def response_body_single_object():
    return """
    {
        "id": "f089354e-8366-4e18-aea3-4cb4a3a50b48",
        "description": "The finance app",
        "name": "Finance",
        "publishedBy": "Bill",
        "lastUpdate": "2019-01-13T09:46:53.094+02:00"
    }
    """


@pytest.fixture
def response_body_multiple_objects():
    return """
    {
        "value": [
            {
            "id": "f089354e-8366-4e18-aea3-4cb4a3a50b48",
            "isReadOnly": false,
            "isOnDedicatedCapacity": false,
            "name": "sample group"
            },
            {
            "id": "3d9b93c6-7b6d-4801-a491-1738910904fd",
            "isReadOnly": false,
            "isOnDedicatedCapacity": false,
            "name": "marketing group"
            },
            {
            "id": "a2f89923-421a-464e-bf4c-25eab39bb09f",
            "isReadOnly": false,
            "isOnDedicatedCapacity": false,
            "name": "contoso",
            "dataflowStorageId": "d692ae06-708c-485e-9987-06ff0fbdbb1f"
            }
        ]
    }
    """


@responses.activate
def test_get_raw_single_object(
    session,
    response_body_single_object,
):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/apps/f089354e-8366-4e18-aea3-4cb4a3a50b48",
        body=response_body_single_object,
    )

    resource = (
        "https://api.powerbi.com/v1.0/myorg/apps/f089354e-8366-4e18-aea3-4cb4a3a50b48"
    )
    raw = _utils.get_raw(resource, session)

    assert "id" in raw
    assert "description" in raw
    assert "name" in raw
    assert "publishedBy" in raw
    assert "lastUpdate" in raw


@responses.activate
def test_get_error_includes_additional_info(session):
    resource = "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229"

    responses.get(
        resource,
        body='{"error":{"code":"ItemNotFound","message":"Dataset cfafbeb1-8037-4d0c-896e-a46fb27ff229 is not found."}}',
        status=404,
    )

    with pytest.raises(HTTPError) as ex:
        _utils.get(resource, session)

    assert "cfafbeb1-8037-4d0c-896e-a46fb27ff229" in str(ex.value)


@responses.activate
def test_get_raises_no_body(session):
    resource = "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229"

    responses.get(resource, status=404)

    with pytest.raises(HTTPError) as ex:
        _utils.get(resource, session)


@responses.activate
def test_get_raises_with_exception_body(session):
    resource = "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229"

    responses.get(
        resource,
        body=Exception("Something Exceptional happened."),
        status=404,
    )

    with pytest.raises(Exception) as ex:
        _utils.get(resource, session)


@responses.activate
def test_get_raw_single_object(
    session,
    response_body_multiple_objects,
):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/apps/f089354e-8366-4e18-aea3-4cb4a3a50b48",
        body=response_body_multiple_objects,
    )

    resource = (
        "https://api.powerbi.com/v1.0/myorg/apps/f089354e-8366-4e18-aea3-4cb4a3a50b48"
    )
    raw = _utils.get_raw(resource, session)

    assert isinstance(raw, list)


@responses.activate
def test_get_raw_raises(
    session,
):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/apps/f089354e-8366-4e18-aea3-4cb4a3a50b48",
        body="{}",
        status=404,
    )

    resource = (
        "https://api.powerbi.com/v1.0/myorg/apps/f089354e-8366-4e18-aea3-4cb4a3a50b48"
    )

    with pytest.raises(HTTPError):
        _utils.get_raw(resource, session)


@responses.activate
def test_put(session):
    json_params = {
        "identifier": "john@contoso.com",
        "principalType": "User",
        "datasetUserAccessRight": "Read",
    }

    responses.put(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/users",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    resource = "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/users"
    payload = {
        "identifier": "john@contoso.com",
        "principalType": "User",
        "datasetUserAccessRight": "Read",
    }

    _utils.put(resource, session, payload)


@responses.activate
def test_put_raises(session):
    json_params = {
        "identifier": "john@contoso.com",
        "principalType": "User",
        "datasetUserAccessRight": "Read",
    }

    responses.put(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/users",
        match=[matchers.json_params_matcher(json_params)],
        body="{}",
        status=400,
    )

    resource = "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229/users"
    payload = {
        "identifier": "john@contoso.com",
        "principalType": "User",
        "datasetUserAccessRight": "Read",
    }

    with pytest.raises(HTTPError):
        _utils.put(resource, session, payload)


@responses.activate
def test_patch(session):
    json_params = {
        "targetStorageMode": "PremiumFiles",
    }

    responses.patch(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    resource = "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229"
    payload = {
        "targetStorageMode": "PremiumFiles",
    }

    _utils.patch(resource, session, payload)


@responses.activate
def test_request_mixin_patch_raises(session):
    json_params = {
        "targetStorageMode": "PremiumFiles",
    }

    responses.patch(
        "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        match=[
            matchers.json_params_matcher(json_params),
        ],
        body="{}",
        status=400,
    )

    resource = "https://api.powerbi.com/v1.0/myorg/datasets/cfafbeb1-8037-4d0c-896e-a46fb27ff229"
    payload = {
        "targetStorageMode": "PremiumFiles",
    }

    with pytest.raises(HTTPError):
        _utils.patch(resource, session, payload)


def test_build_url():
    path_format = "/reports/{}/datasources"

    id = "cfafbeb1-8037-4d0c-896e-a46fb27ff228"

    path = _utils.build_path(path_format, id)
    expected_path = "/reports/cfafbeb1-8037-4d0c-896e-a46fb27ff228/datasources"

    assert expected_path == path


def test_build_url_with_obj():
    path_format = "/reports/{}/datasources"

    id = Report(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff228",
        session=requests.Session(),
    )

    path = _utils.build_path(path_format, id)
    expected_path = "/reports/cfafbeb1-8037-4d0c-896e-a46fb27ff228/datasources"

    assert expected_path == path


def test_build_url_with_multiple_obj():
    path_format = "/groups/{}/reports/{}/datasources"

    report = Report(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
    )

    group = Group(
        id="f089354e-8366-4e18-aea3-4cb4a3a50b48",
        session=requests.Session(),
    )

    path = _utils.build_path(path_format, group, report)
    expected_path = "/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/reports/cfafbeb1-8037-4d0c-896e-a46fb27ff229/datasources"

    assert expected_path == path



"""Utility functions that are consumed internally by pbipy."""

from pathlib import Path
import re
from typing import TYPE_CHECKING

from requests import RequestException, Response, Session

if TYPE_CHECKING:
    from pbipy.resources import Resource


def build_path(
    path_format: str,
    *identifiers: "str | Resource",
) -> str:
    """
    Build a path to a resource from the supplied components.

    Parameters
    ----------
    `path_format` : `str`
        The resource format with bracketed placeholders, e.g., `reports/{}/datasources`
    `*identifiers` : `str | Resource`
        Ids or Resource Objects with an `id` attribute. The provided Ids
        will be added to the path according `path_format`.

    Returns
    -------
    `str`
        Path to a resource.

    """

    ids = []

    for identifier in identifiers:
        try:
            ids.append(identifier.id)
        except AttributeError:
            ids.append(identifier)

    path = path_format.format(*ids)

    return path


def to_snake_case(s):
    pattern = re.compile("((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))")
    return pattern.sub(r"_\1", s).lower()


def to_camel_case(text):
    s = text.replace("-", " ").replace("_", " ")
    s = s.split()

    if len(text) == 0:
        return text

    return s[0] + "".join(i.capitalize() for i in s[1:])


def remove_no_values(
    d: dict,
) -> dict:
    """
    Recursively remove keys with a value of `None` or `{}` from a dictionary.
    If the result of removing `None` is `{}`, then this is removed until
    only keys with values remain.

    Parameters
    ----------
    `d` : `dict`
        dict to remove None or empty dicts from.

    Returns
    -------
    `dict`
        dict with `None` or `{}` removed.
    """

    new_d = {}

    for k, v in d.items():
        if isinstance(v, dict):
            v = remove_no_values(v)
        if not v in (None, {}):
            new_d[k] = v

    return new_d


def file_path_from_components(
    file_name: str,
    extension: str,
    directory: str | Path = None,
):
    """
    Create a file path from file path components.

    Parameters
    ----------
    `file_name` : `str`
        Name of the file (without extension).
    `extension` : `str`
        The file extension.
    `directory` : `str | Path`, optional
        Directory of the file.

    Returns
    -------
    `Path`
        File path to the file.

    """

    if isinstance(directory, str):
        f_dir = Path(directory)
    elif directory is None:
        f_dir = Path()  # cwd
    else:
        f_dir = directory

    # Be a bit more flexible and accept .ext
    if extension[0] == ".":
        f_ext = extension[1:]
    else:
        f_ext = extension

    return f_dir / f"{file_name}.{f_ext}"


def to_identifier(
    s: str,
) -> str:
    """
    Convert a string into a valid Python identifier, e.g., variable or
    attribute name.

    Parameters
    ----------
    `s` : `str`
        String to convert.

    Returns
    -------
    `str`
        Converted string.

    """

    # Remove leading characters until letter or underscore
    s = re.sub("^[^a-zA-Z_]+", "", s)

    # Replace invalid characters
    s = re.sub("[^0-9a-zA-Z_]", "_", s)

    return s


def parse_raw(
    raw: dict,
) -> dict | list[dict]:
    """
    Parses json from an api response.

    Depending on the endpoint, the json will be a dict representing the
    requested object, or a list of dicts at the `"values"` key. If a list,
    this method will return the list, rather than the original json dict.

    Parameters
    ----------
    `raw` : `dict`
        Raw json from the api response.

    Returns
    -------
    `dict | list[dict]`
        Either dict representation of the object, or a list of dict
        representations.

    """

    if "value" in raw:
        return raw["value"]
    else:
        return raw


def raise_error(
    response: Response,
) -> None:
    """
    Wrapper around request's `raise_for_status()` method. Customizes the
    error message to include any extra information provided by the api.

    Parameters
    ----------
    `response` : `Response`
        Requests `Response` object.

    Raises
    ------
    `Exception`
        If there was an error in the response.

    """

    try:
        js = response.json()
    except Exception:
        js = None

    try:
        response.raise_for_status()
    except RequestException as rex:
        if js:
            if len(rex.args) >= 1:
                rex.args = (
                    rex.args[0] + f". Additional information from API: {js}",
                ) + rex.args[1:]

        raise rex


def get(
    resource: str,
    session: Session,
    params: dict = None,
    **kwargs: dict,
) -> Response:
    """
    Get a resource from an api endpoint. Wraps request's `get` method to
    include api-specific error handling.

    Parameters
    ----------
    `resource` : `str`
        URL of the resource.
    `session` : `Session`
        Authenticated `requests.Session` object used to make the request.
    `params` : `dict`, optional
        Request parameters.

    Returns
    -------
    `Response`
        requests `Response` object.

    Raises
    ------
    `Exception`
        If there was an error during the request process.

    """

    try:
        response = session.get(resource, params=params, **kwargs)
        raise_error(response)
    except Exception as ex:
        raise ex

    return response


def get_raw(
    resource: str,
    session: Session,
    params: dict = None,
    **kwargs: dict,
) -> dict | list:
    """
    Convenience function that makes a get request to an api resource, handles
    the response, and returns parsed json.

    Parameters
    ----------
    `resource` : `str`
        URL of the api resource.
    `session` : `Session`
        Authenticated `requests.Session` object used to make the request.

    Returns
    -------
    `dict | list`
        Parsed response json using `_utils.parse_raw`.

    Raises
    ------
    `Exception`
        If there was an error during the request process.

    """

    try:
        response = get(
            resource,
            session,
            params,
            **kwargs,
        )
        raw = response.json()

        return parse_raw(raw)

    except Exception as ex:
        raise ex


def post(
    resource: str,
    session: Session,
    payload: dict = None,
    params: dict = None,
) -> Response:
    """
    Post data to an api endpoint. Wraps request's `post` method to include
    api-specific error handling.

    Parameters
    ----------
    `resource` : `str`
        URL of the resource to post to.
    `session` : `Session`
        Authenticated `requests.Session` object used to make the request.
    `payload` : `dict`, optional
        Data to post to the resource.
    `params` : `dict`, optional
        Request parameters.

    Returns
    -------
    `Response`
        Response generated from the post request.

    Raises
    ------
    `Exception`
        If there was an error during the request process.

    """

    try:
        response = session.post(
            resource,
            params=params,
            json=payload,
        )
        raise_error(response)
    except Exception as ex:
        raise ex

    return response


def post_raw(
    resource: str,
    session: Session,
    payload: dict = None,
    **kwargs: dict,
) -> dict:
    """
    Convenience function that posts to an api endpoint, handles the response,
    and returns parsed json.

    Parameters
    ----------
    `resource` : `str`
        URL of the resource to post to.
    `session` : `Session`
        Authenticated `requests.Session` object used to make the request.
    `payload` : `dict`, optional
        Request data.

    Returns
    -------
    `dict`
        Parsed response json using `_utils.parse_raw`.

    Raises
    ------
    `Exception`
        If there was an error during the request process.

    """

    try:
        response = post(
            resource,
            session,
            payload,
            **kwargs,
        )
        raw = response.json()

        return parse_raw(raw)

    except Exception as ex:
        raise ex


def put(
    resource: str,
    session: Session,
    payload: dict,
) -> Response:
    """
    Make a put request to an api endpoint. Wraps request's `put` method
    to include api-specific error handling.

    Parameters
    ----------
    `resource` : `str`
        URL of the resource.
    `session` : `Session`
        Authenticated `requests.Session` object used to make the request.
    `payload` : `dict`
        Payload data to include with the put request.

    Returns
    -------
    `Response`
        requests Response object.

    Raises
    ------
    `Exception`
        If there was an error during the request process.

    """

    try:
        response = session.put(resource, json=payload)
        raise_error(response)
    except Exception as ex:
        raise ex

    return response


def patch(
    resource: str,
    session: Session,
    payload: dict = None,
) -> Response:
    """
    Make a patch request to an api endpoint. Wraps request's `patch` method
    to include api-specific error handling.

    Parameters
    ----------
    `resource` : `str`
        URL of the resource to patch.
    `session` : `Session`
        Authenticated `requests.Session` object used to make the request.
    `payload` : `dict`
        Data to patch on the resource.

    Returns
    -------
    `Response`
        Response generated from the patch request.

    Raises
    ------
    `Exception`
        If there was an error during the request process.
    """

    try:
        response = session.patch(
            resource,
            json=payload,
        )
        raise_error(response)
    except Exception as ex:
        raise ex

    return response


def delete(
    resource: str,
    session: Session,
    params: dict = None,
) -> Response:
    """
    Make a delete request to an api endpoint. Wraps request's `delete`
    method to include api-specific error handling.

    Parameters
    ----------
    `resource` : `str`
        URL of the resource.
    `session` : `Session`
        Authenticated `requests.Session` object used to make the request.

    Returns
    -------
    `Response`
        Response generated from the delete request.

    Raises
    ------
    `Exception`
        If there was an error during the request process.

    """

    try:
        response = session.delete(
            resource,
            params=params,
        )
        raise_error(response)
    except Exception as ex:
        raise ex

    return response



"""
Module implements a wrapper around the Power BI Rest API Dataflow operations.

Full Dataflow Operations documentation can be found at:

https://learn.microsoft.com/en-us/rest/api/power-bi/dataflows

"""

from requests import Session

from pbipy.resources import Resource
from pbipy import _utils


class Dataflow(Resource):
    """
    A Power BI Dataflow.

    Users should initialize a `Dataflow` object by calling the `dataflow()` method
    on the `PowerBI` client, rather than creating directly.

    Examples
    --------
    Retrieving a `Dataflow` object using a `pbi` client object.

    ```
    >>> my_dataflow = pbi.dataflow("bd32e5c0-363f-430b-a03b-5535a4804b9b")
    ```

    """

    _REPR = [
        "id",
        "name",
        "description",
        "group_id",
    ]

    def __init__(
        self,
        id: str,
        session: Session,
        group_id: str,
        raw: dict = None,
    ) -> None:
        super().__init__(id, session)

        self.group_id = group_id

        self.resource_path = f"/groups/{self.group_id}/dataflows/{self.id}"
        self.base_path = f"{self.BASE_URL}{self.resource_path}"

        if raw:
            self._load_from_raw(raw)

    def datasources(
        self,
    ) -> list[dict]:
        """
        Return a list of Datasources for the Dataflow.

        Returns
        -------
        `list[dict]`
            The list of Datasources associated with the Dataflow.

        """

        resource = self.base_path + "/datasources"
        raw = _utils.get_raw(
            resource,
            self.session,
        )

        return raw

    def refresh(
        self,
        notify_option: str,
        process_type: str = None,
    ) -> None:
        """
        Trigger a refresh of the Dataflow.

        Parameters
        ----------
        `notify_option` : `str`
            Email notification options. Supported options are: `MailOnFailure`
            or `NoNotification`. `MailOnCompletion` is not supported.
        `process_type` : `str`, optional
            The type of refresh process to use.

        """

        resource = self.base_path + "/refreshes"
        payload = {"notifyOption": notify_option}
        params = {"processType": process_type}

        _utils.post(
            resource,
            self.session,
            payload=payload,
            params=params,
        )

    def transactions(
        self,
    ) -> list[dict]:
        """
        Returns a list of transactions for the Dataflow.

        Returns
        -------
        `list[dict]`
            List of Dataflow transactions.

        """

        resource = self.base_path + "/transactions"
        raw = _utils.get_raw(
            resource,
            self.session,
        )

        return raw

    def update(
        self,
        name: str = None,
        description: str = None,
        allow_native_queries: bool = None,
        compute_engine_behavior: str = None,
    ) -> None:
        """
        Updates Dataflow properties, capabilities, and settings.

        Parameters
        ----------
        `name` : `str`, optional
            The new name for the Dataflow.
        `description` : `str`, optional
            The new description for the Dataflow.
        `allow_native_queries` : `bool`, optional
            Whether to allow native queries.
        `compute_engine_behavior` : `str`, optional
            The behavior of the compute engine.

        Raises
        ------
        `ValueError`
            If no properties to update were provided.

        """

        update_request = {
            "name": name,
            "description": description,
            "allowNativeQueries": allow_native_queries,
            "computeEngineBehavior": compute_engine_behavior,
        }

        request_body = _utils.remove_no_values(update_request)

        if request_body in [None, {}]:
            raise ValueError(
                "No properties were provided to update. Please specify at least one property to update."
            )

        resource = self.base_path

        _utils.patch(
            resource,
            self.session,
            request_body,
        )

    def update_refresh_schedule(
        self,
        notify_option: str = None,
        days: list[str] = None,
        enabled: bool = None,
        local_time_zone_id: str = None,
        times: list[str] = None,
    ) -> None:
        """
        Creates or updates the Refresh Schedule for the Dataflow.

        Parameters
        ----------
        `notify_option` : `str`, optional
            Email notification option, e.g. "MailOnCompletion", "MailOnFailure",
            or "NoNotification".
        `days` : `list[str]`, optional
             The full name of days on which to execute the refresh,e.g, "Monday",
            "Tuesday", "Wednesday", etc.
        `enabled` : `bool`, optional
            Enable/Disable the Refresh Schedule.
        `local_time_zone_id` : `str`, optional
            The ID of the time zone to use, e.g, "UTC".
        `times` : `list[str]`, optional
            The times of day to execute the refresh expressed as hh:mm, e.g.,
            "07:00", "16:00", etc.

        Raises
        ------
        `ValueError`
            If no values are provided to update.

        """

        refresh_schedule_request = {
            "value": {
                "notifyOption": notify_option,
                "days": days,
                "enabled": enabled,
                "localTimeZoneId": local_time_zone_id,
                "times": times,
            }
        }

        request_body = _utils.remove_no_values(refresh_schedule_request)

        if request_body in [None, {}]:
            raise ValueError(
                "No options were provided to update. Please specify an option to update."
            )

        resource = self.base_path + "/refreshSchedule"

        _utils.patch(
            resource,
            self.session,
            request_body,
        )

    def upstream_dataflows(
        self,
    ) -> list[dict]:
        """
        Returns a list of upstream dataflows for the Dataflow.

        Returns
        -------
        `list[dict]`
            List of Dependent Dataflows.

        """

        resource = self.base_path + "/upstreamDataflows"
        raw = _utils.get_raw(
            resource,
            self.session,
        )

        return raw



"""
Module implements a Power BI client that wraps around the Power BI Rest 
API.

Full API documentation can be found at:

https://learn.microsoft.com/en-us/rest/api/power-bi/

"""

from pathlib import Path
from typing import IO

import requests

from pbipy import settings
from pbipy.admin import Admin
from pbipy.apps import App
from pbipy.dashboards import Dashboard
from pbipy.dataflows import Dataflow
from pbipy.datasets import Dataset
from pbipy.groups import Group
from pbipy.imports import Import, TemporaryUploadLocation
from pbipy.reports import Report
from pbipy import _utils


class PowerBI:
    """
    User Interface into Power BI Rest Api.

    Users interact with their Power BI service by constructing an instance
    of this object and calling its methods.

    Authentication with the Power BI service requires a `bearer_token` which
    must be generated in advance of initializing the `PowerBI` client. How
    the token is generated depends on the user's Azure and Power BI configuration.

    In general, `PowerBI()` methods wrap the operations described here:

    https://learn.microsoft.com/en-us/rest/api/power-bi/

    Parameters
    ----------
    `bearer_token` : `str`
        Bearer token used to authenticate with your Power BI service.
    `session` : `requests.Session`, optional
        `Session` object used to make http requests. Users can subclass
        a `Session` and pass to the constructor of the client to implement
        customized request handling, e.g., implementing a retry strategy.

    Examples
    --------
    Initializing the client.

    ```
    >>> from powerbi import PowerBI
    >>> pbi = PowerBI(bearer_token="aBCdEFGhijK123456xYz")
    ```

    Using the client to retrieve a report and then initiate a refresh.

    ```
    >>> my_report = pbi.report("5b218778-e7a5-4d73-8187-f10824047715")
    >>> my_report.refresh()
    ```

    Using the client to create a new Workspace (Group) in the user's Power
    BI instance.

    ```
    >>> my_workspace = pbi.create_group("Workspace Name")
    ```

    """

    BASE_URL = settings.BASE_URL

    def __init__(
        self,
        bearer_token: str,
        session: requests.Session = None,
    ) -> None:
        self.bearer_token = bearer_token

        if not session:
            self.session = requests.Session()

        self.session.headers.update({"Authorization": f"Bearer {self.bearer_token}"})

    def admin(
        self,
    ) -> Admin:
        """
        Initialize and return an `Admin` object. The `Admin` object is used
        to access admin-only functionality.

        Returns
        -------
        `Admin`
            The initialized `Admin` object.

        """

        return Admin(self.session)

    def app(
        self,
        app: str,
    ) -> App:
        """
        Return the specified installed app.

        Parameters
        ----------
        `app` : `str`
            App id of the App to retrieve.

        Returns
        -------
        `App`
            The specified App.

        """

        if isinstance(app, App):
            return app

        app = App(app, self.session)
        app.load()

        return app

    def apps(
        self,
    ) -> list[App]:
        resource = self.BASE_URL + "/apps"
        raw = _utils.get_raw(
            resource,
            self.session,
        )

        apps = [
            App(
                app_js.get("id"),
                self.session,
                raw=app_js,
            )
            for app_js in raw
        ]

        return apps

    def cancel_transaction(
        self,
        transaction_id: str,
        group: str | Group,
    ) -> dict:
        """
        Attempt to cancel the specified Dataflow Transaction.

        Parameters
        ----------
        `transaction_id` : `str`
            Id of the Transaction to cancel.
        `group` : `str | Group`
            Group Id or `Group` object where the Transaction originated.
            Should match the Group of the Dataflow that generated the
            transaction.

        Returns
        -------
        `dict`
            Dataflow transaction status.

        """

        if isinstance(group, Group):
            group_id = group.id
        else:
            group_id = group

        resource = (
            self.BASE_URL
            + f"/groups/{group_id}/dataflows/transactions/{transaction_id}/cancel"
        )
        raw = _utils.post_raw(
            resource,
            self.session,
        )

        return raw

    def add_dashboard(
        self,
        name: str,
        group: str | Group = None,
    ) -> Dashboard:
        """
        Creates a new empty dashboard in current users workspace or the
        specified Workspace.

        Parameters
        ----------
        `name` : `str`
            The name of the new dashboard.
        `group` : `str | Group`, optional
            The Group Id or `Group` object to target.

        Returns
        -------
        `Dashboard`
            The newly created `Dashboard`.

        """

        if isinstance(group, Group):
            group_id = group.id
        else:
            group_id = group

        if group:
            path = _utils.build_path("/groups/{}/dashboards", group)
        else:
            path = "/dashboards"

        url = self.BASE_URL + path
        payload = {"name": name}
        raw = _utils.post_raw(
            url,
            self.session,
            payload=payload,
        )

        dashboard = Dashboard(
            raw.get("id"),
            self.session,
            group_id=group_id,
            raw=raw,
        )

        return dashboard

    def dashboard(
        self,
        dashboard: str | Dashboard,
        group: str | Group = None,
    ) -> Dashboard:
        """
        Returns the specified dashboard from the current user's Workspace
        or the specified Workspace.

        Parameters
        ----------
        `dashboard` : `str | Dashboard`
            The id of the Dashboard to retrieve.
        `group` : `str | Group`, optional
            The Group Id or `Group` object to target.

        Returns
        -------
        `Dashboard`
            The specified Dashboard.

        """

        if isinstance(dashboard, Dashboard):
            return dashboard

        if isinstance(group, Group):
            group_id = group.id
        else:
            group_id = group

        if group:
            path = _utils.build_path("/groups/{}/dashboards/{}", group, dashboard)
        else:
            path = _utils.build_path("/dashboards/{}", dashboard)

        url = self.BASE_URL + path
        raw = _utils.get_raw(
            url,
            self.session,
        )

        dashboard = Dashboard(
            raw.get("id"),
            self.session,
            group_id=group_id,
            raw=raw,
        )

        return dashboard

    def dataflow(
        self,
        dataflow: str | Dataflow,
        group: str | Group,
    ) -> Dataflow:
        """
        Get and load the specified dataflow.

        Parameters
        ----------
        `dataflow` : `str | Dataflow`
            The Dataflow Id of the Dataflow to retrieve.
        `group` : `str | Group`
            The Group Id or `Group` object where the Dataflow resides.

        Returns
        -------
        `Dataflow`
            The specified Dataflow.

        Notes
        -----
        The Get Dataflow endpoint returns a json file. For consistency
        with the rest of pbipy, this method returns a `Dataflow` object instead
        of the file itself. To access the json file, use the `Dataflow.raw`
        property.

        """

        if isinstance(dataflow, Dataflow):
            return Dataflow

        if isinstance(group, Group):
            group_id = group.id
        else:
            group_id = group

        resource = self.BASE_URL + f"/groups/{group_id}/dataflows/{dataflow}"
        raw = _utils.get_raw(
            resource,
            self.session,
        )

        dataflow = Dataflow(
            raw.get("objectId", dataflow),
            self.session,
            group_id=group_id,
            raw=raw,
        )

        return dataflow

    def dataflows(
        self,
        group: str | Group,
    ) -> list[Dataflow]:
        """
        Returns a list of all dataflows from the specified Workspace.

        Parameters
        ----------
        `group` : `str | Group`
            Group Id or `Group` object where the Dataflow resides.

        Returns
        -------
        `list[Dataflow]`
            List of `Dataflow` objects from the specified Workspace.

        """

        if isinstance(group, Group):
            group_id = group.id
        else:
            group_id = group

        resource = self.BASE_URL + f"/groups/{group_id}/dataflows"
        raw = _utils.get_raw(
            resource,
            self.session,
        )

        dataflows = [
            Dataflow(
                dataflow_js.get("objectId"),
                self.session,
                group_id=group_id,
                raw=dataflow_js,
            )
            for dataflow_js in raw
        ]

        return dataflows

    def delete_dataflow(
        self,
        dataflow: str | Dataflow,
        group: str | Group,
    ) -> None:
        """
        Deletes a dataflow from Power BI data prep storage, including its
        definition file and model.

        Parameters
        ----------
        `dataflow` : `str | Dataflow`
            Dataflow Id or `Dataflow` object to delete.
        `group` : `str | Group`
            Group Id or `Group` object where the Dataflow resides.

        """

        if isinstance(dataflow, Dataflow):
            dataflow_id = dataflow.id
        else:
            dataflow_id = dataflow

        if isinstance(group, Group):
            group_id = group.id
        else:
            group_id = group

        resource = self.BASE_URL + f"/groups/{group_id}/dataflows/{dataflow_id}"

        _utils.delete(
            resource,
            self.session,
        )

    # TODO: Add support for passing in a group obj
    def dataset(
        self,
        dataset: str | Dataset,
        group: str = None,
    ) -> Dataset:
        """
        Return the specified dataset from MyWorkspace or the specified
        group.

        Parameters
        ----------
        `dataset` : `str | Dataset`
            Dataset Id of the dataset to retrieve.
        `group` : `str`, optional
            Group Id where the dataset resides., by default None

        Returns
        -------
        `Dataset`
            The specified dataset.

        """

        if isinstance(dataset, Dataset):
            return dataset

        dataset = Dataset(dataset, self.session, group_id=group)
        dataset.load()

        return dataset

    def datasets(
        self,
        group: str = None,
    ) -> list[Dataset]:
        """
        Returns a list of datasets from MyWorkspace or the specified group.

        Parameters
        ----------
        `group` : str, optional
            Group Id where the datasets reside. If not supplied, then datasets
            will be retrieved from MyWorkspace.

        Returns
        -------
        `list[Dataset]`
            List of datasets for MyWorkspace or the specified group.

        """

        if group:
            path = f"/groups/{group}/datasets"
        else:
            path = "/datasets"

        resource = self.BASE_URL + path
        raw = _utils.get_raw(
            resource,
            self.session,
        )

        datasets = [
            Dataset(
                dataset_js.get("id"),
                self.session,
                group_id=group,
                raw=dataset_js,
            )
            for dataset_js in raw
        ]

        return datasets

    def delete_dataset(
        self,
        dataset: str | Dataset,
        group: str = None,
    ) -> None:
        """
        Delete the specified dataset from MyWorkspace or the specified
        group.

        Parameters
        ----------
        `dataset` : `str | Dataset`
            Dataset Id or `Dataset` object to delete.
        `group` : `str`, optional
            Group Id where the dataset resides., by default None

        Raises
        ------
        `HTTPError`
            If the api response status code is not equal to 200.

        """
        # FIXME: Doesn't use the group from the Dataset.

        if isinstance(dataset, Dataset):
            dataset_id = dataset.id
        else:
            dataset_id = dataset

        if group:
            path = f"/groups/{group}/datasets/{dataset_id}"
        else:
            path = f"/datasets/{dataset_id}"

        resource = self.BASE_URL + path

        _utils.delete(
            resource,
            self.session,
        )

    def group(
        self,
        group_id: str,
    ) -> Group:
        """
        Return the specified group.

        Convenience function that is the equivalent of
        `group(filter="id eq 'group_id'")`.

        Parameters
        ----------
        `group_id` : `str`
            Group Id of the group to retrieve.

        Returns
        -------
        `Group`
            The specified group.

        Raises
        ------
        `ValueError`
            If the Group was not found by the api.

        """

        id_filter = f"id eq '{group_id}'"

        groups = self.groups(filter=id_filter)

        if groups == []:
            raise ValueError(f"Group Id: {id_filter}, was not found by the API.")

        return groups[0]

    def groups(
        self,
        filter: str = None,
        skip: int = None,
        top: int = None,
    ) -> list[Group]:
        """
        Returns a list of workspaces the user has access to.

        Parameters
        ----------
        `filter` : `str`, optional
            Filters the results, based on a boolean condition, e.g.,
            `contains(name, 'marketing')`, `name eq 'contoso'`
        `skip` : `int`, optional
            Skips the first n results.
        `top` : `int`, optional
            Returns only the first n results.

        Returns
        -------
        `list[Group]`
            List of `Group` objects the user has access to, and/or
            that matched the specified filters.

        Notes
        -----
        See below for more details on filter syntax:
        https://learn.microsoft.com/en-us/power-bi/collaborate-share/service-url-filters

        """

        params = {
            "$filter": filter,
            "$skip": skip,
            "$top": top,
        }

        preparred_params = _utils.remove_no_values(params)

        path = self.BASE_URL + "/groups"
        raw = _utils.get_raw(
            path,
            self.session,
            params=preparred_params,
        )

        groups = [
            Group(
                group_js.get("id"),
                self.session,
                raw=group_js,
            )
            for group_js in raw
        ]

        return groups

    def create_group(
        self,
        name: str,
        workspace_v2: bool = None,
    ) -> Group:
        """
        Create a new workspace.

        Parameters
        ----------
        `name` : `str`
            The name for the new workspace.
        `workspace_v2` : `bool`, optional
            (Preview feature) Whether to create a workspace. The only supported value is `true`.

        Returns
        -------
        `Group`
            The newly created workspace (group).

        """

        payload = {"name": name}

        if workspace_v2 is not None:
            resource = self.BASE_URL + f"/groups?workspaceV2={workspace_v2}"
        else:
            resource = self.BASE_URL + "/groups"

        raw = _utils.post_raw(
            resource,
            self.session,
            payload,
        )

        return Group(
            raw.get("id"),
            self.session,
            raw=raw,
        )

    def delete_group(
        self,
        group: str | Group,
    ) -> None:
        """
        Delete the specified workgroup.

        Parameters
        ----------
        `group` : `str | Group`
            Group Id or `Group` object to delete.

        """

        if isinstance(group, Group):
            group_id = group.id
        else:
            group_id = group

        resource = self.BASE_URL + f"/groups/{group_id}"

        _utils.delete(
            resource,
            self.session,
        )

    def report(
        self,
        report: str | Report,
        group: str | Group = None,
    ) -> Report:
        """
        Return the specified report from MyWorkspace or the specified group.

        Parameters
        ----------
        report : `str | Report`
            Report Id to retrieve.
        group : `str | Group`, optional
            Group Id or `Group` object where the report resides.

        Returns
        -------
        `Report`
            The specified `Report` object.

        """

        if isinstance(report, Report):
            return report

        if isinstance(group, Group):
            group_id = group.id
        else:
            group_id = group

        report = Report(report, self.session, group_id=group_id)
        report.load()

        return report

    def imported_file(
        self,
        import_id: str | Import,
        group: str | Group = None,
    ) -> Import:
        """
        Return details of the specified import (an imported file) from MyWorkspace
        or the specified group. Does not return the imported file itself.

        Parameters
        ----------
        `import_id` : `str | Import`
            Import Id to retrieve
        `group` : `str | Group`, optional
            Group Id or `Group` object where the import resides.

        Returns
        -------
        `Import`
            The specified `Import` object.

        """

        if isinstance(import_id, Import):
            return import_id

        if isinstance(group, Group):
            group_id = group.id
        else:
            group_id = group

        imported_file = Import(
            import_id,
            self.session,
            group_id=group_id,
        )
        imported_file.load()

        return imported_file

    def imported_files(
        self,
        group: str | Group = None,
    ) -> list[Import]:
        """
        Return a list of imports (imported files) from MyWorkspace or the specified group
        (workspace).

        Parameters
        ----------
        `group` : `str | Group`, optional
            Group Id or `Group` object where the imports reside.

        Returns
        -------
        `list[Import]`
            List of `Import` objects from MyWorkspace or the specified group
            (workspace).

        """

        if isinstance(group, Group):
            group_id = group.id
        else:
            group_id = group

        if group_id:
            path = f"/groups/{group_id}/imports"
        else:
            path = "/imports"

        resource = self.BASE_URL + path
        raw = _utils.get_raw(
            resource,
            self.session,
        )

        imports = [
            Import(
                import_js.get("id"),
                self.session,
                group_id=group_id,
                raw=import_js,
            )
            for import_js in raw
        ]

        return imports

    def create_temporary_upload_location(
        self,
        group: str | Group = None,
    ) -> TemporaryUploadLocation:
        """
        Creates a temporary blob storage upload location for importing large
        Power BI `.pbix` files that are between 1 GB and 10 GB in size.

        Parameters
        ----------
        `group` : `str | Group`, optional
            Group Id or `Group` object to associate with the location.

        Returns
        -------
        `TemporaryUploadLocation`
            Temporary blob storage upload location.

        """

        if isinstance(group, Group):
            group_id = group.id
        else:
            group_id = group

        if group_id:
            path = f"/groups/{group_id}/imports/createTemporaryUploadLocation"
        else:
            path = "/imports/createTemporaryUploadLocation"

        resource = self.BASE_URL + path
        raw = _utils.post_raw(
            resource,
            self.session,
        )

        temporary_upload_location = TemporaryUploadLocation(
            expiration_time=raw.get("expirationTime"),
            url=raw.get("url"),
        )

        return temporary_upload_location

    def import_file(
        self,
        filepath_or_filelike: str | Path | IO,
        dataset_display_name: str,
        name_conflict: str = None,
        override_model_label: bool = None,
        override_report_label: bool = None,
        skip_report: bool = None,
        subfolder_object_id: str = None,
        group: str | Group = None,
    ) -> Import:
        """
        Import a file, or file-like object, into MyWorkspace or the specified
        Group, and return an `Import` object representing the newly imported
        file.

        For file sizes between 1GB and 10GB, see the `import_large_file`
        method.

        Parameters
        ----------
        `filepath_or_filelike` : `str | Path | IO`
            File path or file-like object to import.
        `dataset_display_name` : `str`
            The display name of the file, should include file extension.
        `name_conflict` : `str`, optional
            Specifies what to do if file already exists. Available values:
            `Ignore` (default), `Abort`, or `Overwrite`
        `override_model_label` : `bool`, optional
            Whether to override the existing label on a model when republishing
            a Power BI `.pbix` file. The service default value is `true`.
        `override_report_label` : `bool`, optional
            Whether to override the existing report label when republishing
            a Power BI `.pbix` file. The service default value is `true`.
        `skip_report` : `bool`, optional
            Whether to skip report import. If specified, the value must
            be true. Only supported for Power BI `.pbix` files.
        `subfolder_object_id` : `str`, optional
            The subfolder ID to import the file to subfolder.
        `group` : `str | Group`, optional
            Group Id or `Group` object to import the file into.

        Returns
        -------
        `Import`
            `Import` object representing the newly imported file.

        Raises
        ------
        `Exception`
            If an error was encountered during the import process.
        
        See Also
        --------
        `PowerBI.import_large_file`

        Notes
        -----
        * Following a successful upload, the method will query the Get Import
        endpoint to retrieve the complete details of the newly imported file.

        * Method does not currently support importing an .`xlsx` file from
        OneDrive for Business.

        """

        if isinstance(group, Group):
            group_id = group.id
        else:
            group_id = group

        if group_id:
            path = f"/groups/{group_id}/imports"
        else:
            path = "/imports"

        resource = self.BASE_URL + path

        params = {
            "datasetDisplayName": dataset_display_name,
            "nameConflict": name_conflict,
            "overrideModelLabel": override_model_label,
            "overrideReportLabel": override_report_label,
            "skipReport": skip_report,
            "subfolderObjectId": subfolder_object_id,
        }

        if isinstance(filepath_or_filelike, str):
            filepath_or_filelike = Path(filepath_or_filelike)

        try:
            with open(filepath_or_filelike, "rb") as file_contents:
                response = self.session.post(
                    resource,
                    files={"file": file_contents},
                    params=params,
                )

        except TypeError:
            response = self.session.post(
                resource,
                files={"file": filepath_or_filelike},
                params=params,
            )

        _utils.raise_error(response)
        raw = _utils.parse_raw(response.json())

        import_id = raw.get("id")

        return self.imported_file(import_id, group=group_id)

    def import_large_file(
        self,
        filepath_or_filelike: str | Path | IO,
        dataset_display_name: str,
        name_conflict: str = None,
        override_model_label: bool = None,
        override_report_label: bool = None,
        skip_report: bool = None,
        subfolder_object_id: str = None,
        group: str | Group = None,
    ) -> Import:
        """
        Import a large (between 1GB and 10GB) file, or file-like object, into
        MyWorkspace or the specified Group, and return an `Import` object
        representing the newly imported file.

        Convenience function that handles the process of importing large
        files.

        Parameters
        ----------
        `filepath_or_filelike` : `str | Path | IO`
            File path or file-like object to import.
        `dataset_display_name` : `str`
            The display name of the file, should include file extension.
        `name_conflict` : `str`, optional
            Specifies what to do if file already exists. Available values:
            `Ignore` (default), `Abort`, or `Overwrite`
        `override_model_label` : `bool`, optional
            Whether to override the existing label on a model when republishing
            a Power BI `.pbix` file. The service default value is `true`.
        `override_report_label` : `bool`, optional
            Whether to override the existing report label when republishing
            a Power BI `.pbix` file. The service default value is `true`.
        `skip_report` : `bool`, optional
            Whether to skip report import. If specified, the value must be
            true. Only supported for Power BI `.pbix` files.
        `subfolder_object_id` : `str`, optional
            The subfolder ID to import the file to subfolder.
        `group` : `str | Group`, optional
            Group Id or `Group` object to import the file into.

        Returns
        -------
        `Import`
            `Import` object representing the newly imported file.

        Raises
        ------
        `Exception`
            If an error was encountered during the import process.

        Notes
        -----
        * This method uses the Create Temporary Upload endpoint to create
        a temporary blob storage location and upload the file there. The
        shared access signature url is then be supplied to the Post Import
        endpoint, following the large Power BI `.pbix` process outlined here:
        https://learn.microsoft.com/en-us/rest/api/power-bi/imports/create-temporary-upload-location

        * Following a successful upload, the method will query the Get Import
        endpoint to retrieve the complete details of the newly imported file.

        * Method does not currently support importing an .`xlsx` file from
        OneDrive for Business.

        """

        if isinstance(group, Group):
            group_id = group.id
        else:
            group_id = group

        if group_id:
            path = f"/groups/{group_id}/imports"
        else:
            path = "/imports"

        params = {
            "datasetDisplayName": dataset_display_name,
            "nameConflict": name_conflict,
            "overrideModelLabel": override_model_label,
            "overrideReportLabel": override_report_label,
            "skipReport": skip_report,
            "subfolderObjectId": subfolder_object_id,
        }

        if isinstance(filepath_or_filelike, str):
            filepath_or_filelike = Path(filepath_or_filelike)

        temporary_upload_location = self.create_temporary_upload_location(group_id)

        try:
            with open(filepath_or_filelike, "rb") as file_contents:
                response = self.session.post(
                    temporary_upload_location.url,
                    files={"file": file_contents},
                )

        except TypeError:
            response = self.session.post(
                temporary_upload_location.url,
                files={"file": filepath_or_filelike},
            )

        _utils.raise_error(response)

        resource = self.BASE_URL + path
        payload = {"fileUrl": temporary_upload_location.url}

        raw = _utils.post_raw(
            resource,
            self.session,
            params=params,
            payload=payload,
        )

        import_id = raw.get("id")

        return self.imported_file(import_id, group=group_id)

    def reports(
        self,
        group: str | Group = None,
    ) -> list[dict]:
        """
        Return a list of reports for MyWorkspace or the specified
        group (workspace).

        Parameters
        ----------
        `group` : `str | Group`, optional
            Group Id or `Group` object where the reports reside., by default None

        Returns
        -------
        `list[dict]`
            List of `Report` objects for MyWorkspace or the specified
            group (workspace).

        """

        if isinstance(group, Group):
            group_id = group.id
        else:
            group_id = group

        if group_id:
            path = f"/groups/{group_id}/reports"
        else:
            path = "/reports"

        resource = self.BASE_URL + path
        raw = _utils.get_raw(
            resource,
            self.session,
        )

        reports = [
            Report(
                report_js.get("id"),
                self.session,
                group_id=group_id,
                raw=report_js,
            )
            for report_js in raw
        ]

        return reports

    def delete_report(
        self,
        report: str | Report,
        group: str | Group = None,
    ) -> None:
        """
        Delete the specified report.

        If a `Report` type is provided as `report`, then it's `group_id`
        will be taken as the `group` value.

        Parameters
        ----------
        `report` : `str | Report`
            Report Id or `Report` object to delete.
        `group` : `str | Group`, optional
            The Group Id or `Group` object where the report resides. If
            a `Report` object is provided then the `group_id` of that report
            will override the provided group value.

            If no group value is provided, it is assumed the report resides
            in the current user's workspace.

        """

        if isinstance(report, Report):
            report_id = report.id
        else:
            report_id = report

        # If we got report, use its group_id and ignore provided
        if isinstance(report, Report):
            group_id = report.group_id
        else:
            if isinstance(group, Group):
                group_id = group.id
            else:
                group_id = group

        if group_id:
            resource = self.BASE_URL + f"/groups/{group_id}/reports/{report_id}"
        else:
            resource = self.BASE_URL + f"/reports/{report_id}"

        _utils.delete(
            resource,
            self.session,
        )



__title__ = "pbipy"
__description__ = "A Python Library for interacting with the Power BI Rest API."
__url__ = "https://github.com/andrewvillazon/pbipy"
__version__ = "2.10.0"
__author__ = "Andrew Villazon"
__author_email__ = "andrew.villazon@gmail.com"
__license__ = "Apache 2.0"
__copyright__ = "Copyright 2024 Andrew Villazon"


from io import BytesIO
from pathlib import Path
from unittest.mock import mock_open, patch

import pytest
import requests
import responses
from responses import matchers

from pbipy.embedtokens import EmbedToken
from pbipy.groups import Group
from pbipy.reports import Report


@pytest.fixture
def report_with_group():
    raw = {
        "datasetId": "cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        "id": "cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        "name": "SalesMarketing",
        "webUrl": "https://app.powerbi.com//reports/5b218778-e7a5-4d73-8187-f10824047715",
        "embedUrl": "https://app.powerbi.com/reportEmbed?reportId=5b218778-e7a5-4d73-8187-f10824047715",
    }

    report = Report(
        id=raw["id"],
        group_id="f089354e-8366-4e18-aea3-4cb4a3a50b48",
        session=requests.Session(),
        raw=raw,
    )

    return report


@pytest.fixture
def report():
    raw = {
        "id": "879445d6-3a9e-4a74-b5ae-7c0ddabf0f11",
        "name": "SalesMarketing",
        "reportType": "PowerBIReport",
    }

    report = Report(
        id=raw["id"],
        session=requests.Session(),
        raw=raw,
    )

    return report


def test_report_creation_from_raw():
    raw = {
        "datasetId": "cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        "id": "5b218778-e7a5-4d73-8187-f10824047715",
        "name": "SalesMarketing",
        "webUrl": "https://app.powerbi.com/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/reports/5b218778-e7a5-4d73-8187-f10824047715",
        "embedUrl": "https://app.powerbi.com/reportEmbed?reportId=5b218778-e7a5-4d73-8187-f10824047715&groupId=f089354e-8366-4e18-aea3-4cb4a3a50b48",
    }

    report = Report(raw.get("id"), session=requests.Session(), raw=raw)

    assert report.id == "5b218778-e7a5-4d73-8187-f10824047715"
    assert report.name == "SalesMarketing"


@responses.activate
def test_datasources(report_with_group, get_datasources_in_group):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/reports/cfafbeb1-8037-4d0c-896e-a46fb27ff229/datasources",
        body=get_datasources_in_group,
        content_type="application/json",
    )

    datasources = report_with_group.datasources()

    assert isinstance(datasources, list)
    assert all(isinstance(datasource, dict) for datasource in datasources)
    assert len(datasources) == 1
    assert datasources[0]["datasourceId"] == "f8c56590-43cb-43bf-8daa-233ba2520f55"


@responses.activate
def test_page(get_page):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/reports/879445d6-3a9e-4a74-b5ae-7c0ddabf0f11/pages/ReportSection",
        body=get_page,
        content_type="application/json",
    )

    report = Report(
        id="879445d6-3a9e-4a74-b5ae-7c0ddabf0f11",
        session=requests.Session(),
    )

    page = report.page("ReportSection")

    assert isinstance(page, dict)
    assert page["displayName"] == "Regional Sales Analysis"


@responses.activate
def test_pages(report, get_pages):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/reports/879445d6-3a9e-4a74-b5ae-7c0ddabf0f11/pages",
        body=get_pages,
        content_type="application/json",
    )

    pages = report.pages()

    assert isinstance(pages, list)
    assert all(isinstance(page, dict) for page in pages)
    assert len(pages) == 2


@responses.activate
def test_rebind_call(report):
    json_params = {"datasetId": "cfafbeb1-8037-4d0c-896e-a46fb27ff229"}

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/reports/879445d6-3a9e-4a74-b5ae-7c0ddabf0f11/Rebind",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    report.rebind("cfafbeb1-8037-4d0c-896e-a46fb27ff229")


@responses.activate
def test_take_over_call(report_with_group):
    responses.post(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/reports/cfafbeb1-8037-4d0c-896e-a46fb27ff229/Default.TakeOver"
    )

    report_with_group.take_over()


def test_take_over_raises_type_error(report):
    with pytest.raises(TypeError):
        report.take_over()


@responses.activate
def test_update_datasources_call_single(report):
    json_params = [
        {
            "datasourceName": "SqlDatasource",
            "connectionDetails": {
                "server": "New-Sql-Server",
                "database": "New-Sql-Database",
            },
        },
    ]

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/reports/879445d6-3a9e-4a74-b5ae-7c0ddabf0f11/Default.UpdateDatasources",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    details = {
        "datasourceName": "SqlDatasource",
        "connectionDetails": {
            "server": "New-Sql-Server",
            "database": "New-Sql-Database",
        },
    }

    report.update_datasources(details)


@responses.activate
def test_update_datasources_call_multiple(report):
    json_params = [
        {
            "datasourceName": "SqlDatasource",
            "connectionDetails": {
                "server": "New-Sql-Server",
                "database": "New-Sql-Database",
            },
        },
        {
            "datasourceName": "SqlAzureDatasource",
            "connectionDetails": {
                "server": "New-SqlAzure-Server.windows.net",
                "database": "New-SqlAzure-Database",
            },
        },
    ]

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/reports/879445d6-3a9e-4a74-b5ae-7c0ddabf0f11/Default.UpdateDatasources",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    details = [
        {
            "datasourceName": "SqlDatasource",
            "connectionDetails": {
                "server": "New-Sql-Server",
                "database": "New-Sql-Database",
            },
        },
        {
            "datasourceName": "SqlAzureDatasource",
            "connectionDetails": {
                "server": "New-SqlAzure-Server.windows.net",
                "database": "New-SqlAzure-Database",
            },
        },
    ]

    report.update_datasources(details)


@responses.activate
def test_update_content_report_object(report):
    json_params = {
        "sourceReport": {
            "sourceReportId": "8e4d5880-81d6-4804-ab97-054665050799",
            "sourceWorkspaceId": "2f42a406-a075-4a15-bbf2-97ef958c94cb",
        },
        "sourceType": "ExistingReport",
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/reports/879445d6-3a9e-4a74-b5ae-7c0ddabf0f11/UpdateReportContent",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    source_report = Report(
        id="8e4d5880-81d6-4804-ab97-054665050799",
        group_id="2f42a406-a075-4a15-bbf2-97ef958c94cb",
        session=requests.Session(),
    )

    report.update_content(source_report)


@responses.activate
def test_update_content_report_no_group(report):
    json_params = {
        "sourceReport": {
            "sourceReportId": "8e4d5880-81d6-4804-ab97-054665050799",
        },
        "sourceType": "ExistingReport",
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/reports/879445d6-3a9e-4a74-b5ae-7c0ddabf0f11/UpdateReportContent",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    source_report = Report(
        id="8e4d5880-81d6-4804-ab97-054665050799",
        session=requests.Session(),
    )

    report.update_content(source_report)


@responses.activate
def test_update_content_report_str_group_object(report):
    json_params = {
        "sourceReport": {
            "sourceReportId": "8e4d5880-81d6-4804-ab97-054665050799",
            "sourceWorkspaceId": "2f42a406-a075-4a15-bbf2-97ef958c94cb",
        },
        "sourceType": "ExistingReport",
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/reports/879445d6-3a9e-4a74-b5ae-7c0ddabf0f11/UpdateReportContent",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    group = Group(
        id="2f42a406-a075-4a15-bbf2-97ef958c94cb",
        session=requests.Session(),
    )

    report.update_content(
        "8e4d5880-81d6-4804-ab97-054665050799",
        group,
    )


@responses.activate
def test_update_content_report_str_group_str(report):
    json_params = {
        "sourceReport": {
            "sourceReportId": "8e4d5880-81d6-4804-ab97-054665050799",
            "sourceWorkspaceId": "2f42a406-a075-4a15-bbf2-97ef958c94cb",
        },
        "sourceType": "ExistingReport",
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/reports/879445d6-3a9e-4a74-b5ae-7c0ddabf0f11/UpdateReportContent",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    report.update_content(
        "8e4d5880-81d6-4804-ab97-054665050799",
        "2f42a406-a075-4a15-bbf2-97ef958c94cb",
    )


@responses.activate
def test_update_content_ignores_group(report):
    # Report has no group, should ignore group id and not pass in as param.
    json_params = {
        "sourceReport": {
            "sourceReportId": "8e4d5880-81d6-4804-ab97-054665050799",
        },
        "sourceType": "ExistingReport",
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/reports/879445d6-3a9e-4a74-b5ae-7c0ddabf0f11/UpdateReportContent",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    source_report = Report(
        id="8e4d5880-81d6-4804-ab97-054665050799",
        session=requests.Session(),
    )

    report.update_content(source_report, "2f42a406-a075-4a15-bbf2-97ef958c94cb")


@responses.activate
def test_update_content_report_str_only_no_group(report):
    json_params = {
        "sourceReport": {
            "sourceReportId": "8e4d5880-81d6-4804-ab97-054665050799",
        },
        "sourceType": "ExistingReport",
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/reports/879445d6-3a9e-4a74-b5ae-7c0ddabf0f11/UpdateReportContent",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    report.update_content("8e4d5880-81d6-4804-ab97-054665050799")


@responses.activate
def test_clone_report_call_with_group(report_with_group):
    json_params = {
        "name": "Attack of the clones",
        "targetWorkspaceId": "2f42a406-a075-4a15-bbf2-97ef958c94cb",
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/reports/cfafbeb1-8037-4d0c-896e-a46fb27ff229/Clone",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    report_with_group.clone(
        "Attack of the clones",
        target_group="2f42a406-a075-4a15-bbf2-97ef958c94cb",
    )


@responses.activate
def test_clone_report_call_with_no_group(report_with_group):
    json_params = {
        "name": "Attack of the clones",
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/reports/cfafbeb1-8037-4d0c-896e-a46fb27ff229/Clone",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    report_with_group.clone("Attack of the clones")


@responses.activate
def test_clone_report_call_with_group_and_dataset(report_with_group):
    json_params = {
        "name": "Attack of the clones",
        "targetModelId": "8e4d5880-81d6-4804-ab97-054665050799",
        "targetWorkspaceId": "2f42a406-a075-4a15-bbf2-97ef958c94cb",
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/reports/cfafbeb1-8037-4d0c-896e-a46fb27ff229/Clone",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    report_with_group.clone(
        "Attack of the clones",
        target_group="2f42a406-a075-4a15-bbf2-97ef958c94cb",
        target_dataset="8e4d5880-81d6-4804-ab97-054665050799",
    )


@patch("pbipy.reports.urlopen")
def test_download(mock_urlopen, report):
    mock_urlopen.return_value = BytesIO()
    mo = mock_open()

    with patch("builtins.open", mo):
        report.download()

    expected_url = "https://api.powerbi.com/v1.0/myorg/reports/879445d6-3a9e-4a74-b5ae-7c0ddabf0f11/Export"
    actual_url = mock_urlopen.call_args.args[0].full_url
    expected_path = Path("SalesMarketing.pbix")

    assert actual_url == expected_url
    mo.assert_called_with(expected_path, "wb")


@patch("pbipy.reports.urlopen")
def test_download_with_dir(mock_urlopen, report):
    mock_urlopen.return_value = BytesIO()
    mo = mock_open()

    with patch("builtins.open", mo):
        report.download(save_to="C:/temp", file_name="NotSalesMarketing")

    expected_url = "https://api.powerbi.com/v1.0/myorg/reports/879445d6-3a9e-4a74-b5ae-7c0ddabf0f11/Export"
    actual_url = mock_urlopen.call_args.args[0].full_url
    expected_path = Path("C:/temp/NotSalesMarketing.pbix")

    assert actual_url == expected_url
    mo.assert_called_with(expected_path, "wb")


@responses.activate
def test_export_request_call(report, export_to_file):
    responses.post(
        "https://api.powerbi.com/v1.0/myorg/reports/879445d6-3a9e-4a74-b5ae-7c0ddabf0f11/ExportTo",
        match=[
            matchers.json_params_matcher(
                {"format": "PDF"},
            ),
        ],
        body=export_to_file,
        status=202,
    )

    report.export_request("pdf")


@responses.activate
def test_export_request_result(report, export_to_file):
    responses.post(
        "https://api.powerbi.com/v1.0/myorg/reports/879445d6-3a9e-4a74-b5ae-7c0ddabf0f11/ExportTo",
        match=[
            matchers.json_params_matcher(
                {"format": "PDF"},
            ),
        ],
        body=export_to_file,
        status=202,
    )

    export_job = report.export_request("pdf")

    assert isinstance(export_job, dict)
    assert export_job.get("percentComplete") == 70
    assert export_job.get("status") == "Running"


@responses.activate
def test_export_status_raw_only(report, get_export_to_file_status):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/reports/879445d6-3a9e-4a74-b5ae-7c0ddabf0f11/exports/Mi9C5419i....PS4=",
        body=get_export_to_file_status,
    )

    status = report.export_status("Mi9C5419i....PS4=")

    assert isinstance(status, dict)
    assert status.get("id") == "Mi9C5419i....PS4="
    assert status.get("status") == "Succeeded"
    assert status.get("percentComplete") == 100


@responses.activate
def test_export_status_retry_after(report, get_export_to_file_status):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/reports/879445d6-3a9e-4a74-b5ae-7c0ddabf0f11/exports/Mi9C5419i....PS4=",
        body=get_export_to_file_status,
        headers={"Retry-After": "5"},
    )

    status = report.export_status("Mi9C5419i....PS4=", include_retry_after=True)

    export_status, retry_after = status

    assert isinstance(status, tuple)
    assert isinstance(export_status, dict)
    assert isinstance(retry_after, int)


@responses.activate
def test_export_status_no_retry_after(report, get_export_to_file_status):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/reports/879445d6-3a9e-4a74-b5ae-7c0ddabf0f11/exports/Mi9C5419i....PS4=",
        body=get_export_to_file_status,
    )

    status = report.export_status("Mi9C5419i....PS4=", include_retry_after=True)

    export_status, retry_after = status

    assert isinstance(status, tuple)
    assert isinstance(export_status, dict)
    assert retry_after is None


@responses.activate
def test_download_export():
    response_file = BytesIO(b"file contents").read()

    responses.get(
        "https://api.powerbi.com/v1.0/myorg/reports/cfafbeb1-8037-4d0c-896e-a46fb27ff229/exports/Mi9C5419i....PS4=/file",
        body=response_file,
        content_type="application/vnd.openxmlformats-officedocument.presentationml.presentation",
    )

    raw = {
        "id": "cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        "name": "SalesMarketing",
        "reportType": "PowerBIReport",
    }

    report = Report(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
        raw=raw,
    )

    m = mock_open()
    with patch("builtins.open", m):
        report.download_export("Mi9C5419i....PS4=")

    m.assert_called_with(Path("SalesMarketing.pptx"), "wb")
    m.return_value.write.assert_called_once_with(response_file)


@responses.activate
def test_download_export_pdf():
    response_file = BytesIO(b"file contents").read()

    responses.get(
        "https://api.powerbi.com/v1.0/myorg/reports/cfafbeb1-8037-4d0c-896e-a46fb27ff229/exports/Mi9C5419i....PS4=/file",
        body=response_file,
        content_type="application/pdf",
    )

    raw = {
        "id": "cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        "name": "SalesMarketing",
        "reportType": "PowerBIReport",
    }

    report = Report(
        id="cfafbeb1-8037-4d0c-896e-a46fb27ff229",
        session=requests.Session(),
        raw=raw,
    )

    m = mock_open()
    with patch("builtins.open", m):
        report.download_export("Mi9C5419i....PS4=", "C:/temp")

    m.assert_called_with(Path("C:/temp/SalesMarketing.pdf"), "wb")
    m.return_value.write.assert_called_once_with(response_file)


@responses.activate
def test_generate_token(
    report_with_group,
    reports_generate_token_in_group,
):
    json_params = {
        "accessLevel": "Edit",
        "allowSaveAs": True,
    }

    responses.post(
        "https://api.powerbi.com/v1.0/myorg/groups/f089354e-8366-4e18-aea3-4cb4a3a50b48/reports/cfafbeb1-8037-4d0c-896e-a46fb27ff229/GenerateToken",
        body=reports_generate_token_in_group,
        content_type="application/json",
        match=[
            matchers.json_params_matcher(json_params),
        ],
    )

    token = report_with_group.generate_token(
        access_level="Edit",
        allow_save_as=True,
    )

    assert isinstance(token, EmbedToken)
    assert token.token == "H4sI....AAA="
    assert token.token_id == "49ae3742-54c0-4c29-af52-619ff93b5c80"


def test_generate_token_raises_TypeError(report):
    with pytest.raises(TypeError):
        report.generate_token()



import pytest
import requests
import responses

from pbipy.apps import App
from pbipy.dashboards import Dashboard, Tile
from pbipy.reports import Report


@pytest.fixture
def app():
    raw = {
        "id": "f089354e-8366-4e18-aea3-4cb4a3a50b48",
        "description": "The finance app",
        "name": "Finance",
        "publishedBy": "Bill",
        "lastUpdate": "2019-01-13T09:46:53.094+02:00",
    }

    app = App(
        id=raw["id"],
        session=requests.Session(),
        raw=raw,
    )

    return app


def test_app_creation(app):
    assert isinstance(app, App)
    assert app.id == "f089354e-8366-4e18-aea3-4cb4a3a50b48"
    assert app.description == "The finance app"


@responses.activate
def test_report(app_get_report, app):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/apps/f089354e-8366-4e18-aea3-4cb4a3a50b48/reports/66b2570c-d9d3-40b2-83d9-1095c6700041",
        body=app_get_report,
        content_type="application/json",
    )

    report = app.report("66b2570c-d9d3-40b2-83d9-1095c6700041")

    assert isinstance(report, Report)
    assert report.app_id == "3d9b93c6-7b6d-4801-a491-1738910904fd"


@responses.activate
def test_reports(app_get_reports, app):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/apps/f089354e-8366-4e18-aea3-4cb4a3a50b48/reports",
        body=app_get_reports,
        content_type="application/json",
    )

    reports = app.reports()

    assert isinstance(reports, list)
    assert all(isinstance(report, Report) for report in reports)
    assert len(reports) == 1
    assert all(hasattr(report, "app_id") for report in reports)


@responses.activate
def test_dashboard(app_get_dashboard, app):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/apps/f089354e-8366-4e18-aea3-4cb4a3a50b48/dashboards/03dac094-2ff8-47e8-b2b9-dedbbc4d22ac",
        body=app_get_dashboard,
        content_type="application/json",
    )

    dashboard = app.dashboard("03dac094-2ff8-47e8-b2b9-dedbbc4d22ac")

    assert isinstance(dashboard, Dashboard)
    assert dashboard.app_id == "3d9b93c6-7b6d-4801-a491-1738910904fd"


@responses.activate
def test_dashboards(app_get_dashboards, app):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/apps/f089354e-8366-4e18-aea3-4cb4a3a50b48/dashboards",
        body=app_get_dashboards,
        content_type="application/json",
    )

    dashboards = app.dashboards()

    assert isinstance(dashboards, list)
    assert all(isinstance(dashboard, Dashboard) for dashboard in dashboards)
    assert all(hasattr(dashboard, "app_id") for dashboard in dashboards)


@responses.activate
def test_tile(app, app_get_tile):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/apps/f089354e-8366-4e18-aea3-4cb4a3a50b48/dashboards/3d9b93c6-7b6d-4801-a491-1738910904fd/tiles/312fbfe9-2eda-44e0-9ed0-ab5dc571bb4b",
        body=app_get_tile,
        content_type="application/json",
    )

    tile = app.tile(
        "312fbfe9-2eda-44e0-9ed0-ab5dc571bb4b",
        dashboard="3d9b93c6-7b6d-4801-a491-1738910904fd",
    )


@responses.activate
def test_tiles(app, app_get_tiles):
    responses.get(
        "https://api.powerbi.com/v1.0/myorg/apps/f089354e-8366-4e18-aea3-4cb4a3a50b48/dashboards/3d9b93c6-7b6d-4801-a491-1738910904fd/tiles",
        body=app_get_tiles,
        content_type="application/json",
    )

    tiles = app.tiles(dashboard="3d9b93c6-7b6d-4801-a491-1738910904fd")

    assert isinstance(tiles, list)
    assert all(isinstance(tile, Tile) for tile in tiles)
    assert len(tiles) == 1



"""
Module implments representations of a Power BI Dashboard and Tile.

These objects are minimal representations and not yet fully implemented.

"""

from requests import Session

from pbipy.resources import Resource


class Dashboard(Resource):
    """A Power BI Dashboard."""
    
    _REPR = [
        "id",
        "display_name",
        "group_id",
        "app_id",
    ]

    def __init__(
        self,
        id: str,
        session: Session,
        group_id=None,
        raw=None,
    ) -> None:
        super().__init__(id, session)

        if group_id:
            self.group_id = group_id
        else:
            self.group_id = None

        if self.group_id:
            self.resource_path = f"/groups/{self.group_id}/dashboards/{self.id}"
        else:
            self.resource_path = f"/dashboards/{self.id}"

        self.base_path = f"{self.BASE_URL}{self.resource_path}"

        if raw:
            self._load_from_raw(raw)


class Tile(Resource):
    """A Power BI Dashboard Tile."""

    _REPR = [
        "id",
        "title",
        "dashboard_id",
        "group_id",
        "report_id",
        "dataset_id",
    ]

    def __init__(
        self,
        id: str,
        dashboard_id: str,
        session: Session,
        group_id=None,
        raw=None,
    ) -> None:
        super().__init__(id, session)

        self.dashboard_id = dashboard_id

        if group_id:
            self.group_id = group_id
        else:
            self.group_id = None

        if self.group_id:
            self.resource_path = f"/groups/{self.group_id}/dashboards/{self.dashboard_id}/tiles/{self.id}"
        else:
            self.resource_path = f"/dashboards/{self.dashboard_id}/tiles/{self.id}"

        if raw:
            self._load_from_raw(raw)



"""
Module implements a wrapper around the Power BI Rest API Admin Operations.

Full Admin Operations documentation can be found at:

https://learn.microsoft.com/en-us/rest/api/power-bi/admin

"""

from datetime import datetime, timezone

import requests

from pbipy import settings
from pbipy.apps import App
from pbipy.dashboards import Dashboard, Tile
from pbipy.dataflows import Dataflow
from pbipy.datasets import Dataset
from pbipy.groups import Group
from pbipy.reports import Report
from pbipy import _utils


class Admin:
    """
    Groups methods that wrap the Power BI Rest API Admin Operations, specialized
    endpoints available to Power BI administrators.

    The `Admin` object should be initialized by calling the `admin()` method
    of the `PowerBI` client object, rather than creating directly.

    Users are free to instantiate the `Admin` object and call its methods,
    but must have administrator rights on their Power BI instance to successfully
    retrieve data from these endpoints.

    """

    BASE_URL = settings.BASE_URL

    def __init__(
        self,
        session: requests.Session,
    ) -> None:
        self.session = session

        self.resource_path = "/admin"
        self.base_path = f"{self.BASE_URL}{self.resource_path}"

    def activity_events(
        self,
        start_date_time: datetime,
        end_date_time: datetime,
        filter: str = None,
    ) -> list[dict]:
        """
        Returns a list of audit Activity Events for the tenant. Activity Events
        are user and admin activities within a Power BI tenant.

        Due to limitations placed on the endpoint, `start_date_time` and
        `end_date_time` must be the same UTC day, i.e. you may only retrieve
        Activity Events one day at a time.

        Parameters
        ----------
        `start_date_time` : `datetime`
            Start date and time of the window for audit event results. Must
            be in ISO 8601 compliant UTC format.
        `end_date_time` : `datetime`
            End date and time of the window for audit event results. Must
            be in ISO 8601 compliant UTC format.
        `filter` : `str`, optional
            Filters the results based on a boolean condition, using `Activity`,
            `UserId`, or both properties. Supports only `eq` and `and` operators.

        Returns
        -------
        `list[dict]`
            List of Activity Events. The structure of each item will depend
            on what type of Activity Event it is.

        Notes
        -----
        As per the design of the associated endpoint, this method will make
        multiple requests as it retrieves the complete Activity Log for the
        specified time window.

        See: https://powerbi.microsoft.com/en-us/blog/the-power-bi-activity-log-makes-it-easy-to-download-activity-data-for-custom-usage-reporting/

        """

        start_iso = start_date_time.isoformat()
        end_iso = end_date_time.isoformat()

        init_params = {
            "startDateTime": f"'{start_iso}'",
            "endDateTime": f"'{end_iso}'",
            "$filter": filter,
        }

        path = "/activityevents"
        url = self.base_path + path

        init_raw = _utils.get_raw(
            url,
            self.session,
            params=init_params,
        )

        activity_events = init_raw["activityEventEntities"]
        continuation_token = init_raw["continuationToken"]

        while continuation_token is not None:
            raw = _utils.get_raw(
                url,
                self.session,
                params={
                    "continuationToken": f"'{continuation_token}'",
                },
            )

            activity_events.extend(raw["activityEventEntities"])
            continuation_token = raw["continuationToken"]

        return activity_events

    def add_encryption_key(
        self,
        name: str,
        key_vault_identifier: str,
        activate: bool,
        is_default: bool,
    ) -> dict:
        """
        Adds an encryption key for Power BI workspaces assigned to a capacity.

        Parameters
        ----------
        `name` : `str`
            The name of the encryption key.
        `key_vault_identifier` : `str`
            The URI that uniquely specifies an encryption key in Azure Key
            Vault.
        `activate` : `bool`
            Whether to activate any inactivated capacities and to use this
            key for its encryption.
        `is_default` : `bool`
            Whether an encryption key is the default key for the entire
            tenant. Any newly created capacity inherits the default key.

        Returns
        -------
        `dict`
            The newly created Tenant Key.

        """

        payload = {
            "name": name,
            "keyVaultKeyIdentifier": key_vault_identifier,
            "activate": activate,
            "isDefault": is_default,
        }

        resource = self.base_path + "/tenantKeys"
        raw = _utils.post_raw(
            resource,
            self.session,
            payload=payload,
        )

        return raw

    def encryption_keys(
        self,
    ) -> list[dict]:
        """
        Returns the Encryption Keys for the tenant.

        Returns
        -------
        `list[dict]`
            List of Encryption Keys for the tenant.

        """

        resource = self.base_path + "/tenantKeys"
        raw = _utils.get_raw(
            resource,
            self.session,
        )

        return raw

    def apps(
        self,
        top: int = None,
    ) -> list[App]:
        """
        Return a list of Apps in the Organization.

        Parameters
        ----------
        `top` : `int`, optional
            The requested number of entries in the refresh history. If not
            provided, the default is all available entries.

        Returns
        -------
        `list[App]`
            The list of Apps for the organization.

        """

        resource = self.base_path + "/apps"
        params = {"$top": top}

        raw = _utils.get_raw(
            resource,
            self.session,
            params,
        )

        apps = [
            App(
                app_js.get("id"),
                self.session,
                raw=app_js,
            )
            for app_js in raw
        ]

        return apps

    def app_users(
        self,
        app: str | App,
    ) -> list[dict]:
        """
        Returns a list of users that have access to the specified app.

        Parameters
        ----------
        `app` : `str | App`
            The App Id or `App` object to target.

        Returns
        -------
        `list[dict]`
            List of Users for the specified app.

        """

        path = _utils.build_path("/apps/{}/users", app)
        url = self.base_path + path
        raw = _utils.get_raw(
            url,
            self.session,
        )

        return raw

    def dashboards(
        self,
        group: str | Group = None,
        expand: str = None,
        filter: str = None,
        skip: int = None,
        top: int = None,
    ) -> list[Dashboard]:
        """
        Returns a list of dashboards for the organization or specified Workspace.

        Parameters
        ----------
        `group` : `str | Group`, optional
            Group Id or `Group` object where the Apps reside. If not supplied
            then returns all apps for the organization.
        `expand` : `str`, optional
            Accepts a comma-separated list of data types, which will be
            expanded inline in the response. Supports tiles.
        `filter` : `str`, optional
            Filters the results, based on a boolean condition.
        `skip` : `int`, optional
            Skips the first n results.
        `top` : `int`, optional
            Returns only the first n results.

        Returns
        -------
        `list[Dashboard]`
            List of `Dashboard` objects for the organization.

        """

        if group:
            path = _utils.build_path("/groups/{}/dashboards", group)
        else:
            path = "/dashboards"

        url = self.base_path + path

        params = {
            "$expand": expand,
            "$filter": filter,
            "$skip": skip,
            "$top": top,
        }

        raw = _utils.get_raw(
            url,
            self.session,
            params=params,
        )

        dashboards = [
            Dashboard(
                dashboard_js.get("id"),
                self.session,
                raw=dashboard_js,
            )
            for dashboard_js in raw
        ]

        return dashboards

    def dashboard_subscriptions(
        self,
        dashboard: str | Dashboard,
    ) -> list[dict]:
        """
        Returns a list of dashboard subscriptions along with subscriber
        details.

        Parameters
        ----------
        `dashboard` : `str | Dashboard`
            Dashboard Id or `Dashboard` object to target.

        Returns
        -------
        `list[dict]`
            List of subscriptions.

        """

        path = _utils.build_path("/dashboards/{}/subscriptions", dashboard)
        url = self.base_path + path
        raw = _utils.get_raw(
            url,
            self.session,
        )

        return raw

    def dashboard_tiles(
        self,
        dashboard: str | Dashboard,
    ) -> list[Tile]:
        """
        Returns a list of tiles within the specified dashboard.

        Parameters
        ----------
        `dashboard` : `str | Dashboard`
            The Dashboard Id or `Dashboard` object to target.

        Returns
        -------
        `list[Tile]`
            List of Tiles from the specified Dashboard.

        """

        path = _utils.build_path("/dashboards/{}/tiles", dashboard)
        url = self.base_path + path
        raw = _utils.get_raw(
            url,
            self.session,
        )

        tiles = [
            Tile(
                tile_js.get("id"),
                dashboard_id=dashboard,
                session=self.session,
                raw=tile_js,
            )
            for tile_js in raw
        ]

        return tiles

    def dashboard_users(
        self,
        dashboard: str | Dashboard,
    ) -> list[dict]:
        """
        Returns a list of users that have access to the specified dashboard.

        Parameters
        ----------
        `dashboard` : `str | Dashboard`
            The Dashboard Id or `Dashboard` object to target.

        Returns
        -------
        `list[dict]`
            List of Dashboard Users.

        """

        path = _utils.build_path("/dashboards/{}/users", dashboard)
        url = self.base_path + path
        raw = _utils.get_raw(
            url,
            self.session,
        )

        return raw

    def dataflow(
        self,
        dataflow: str | Dataflow,
    ) -> Dataflow:
        """
        Get and load the specified Dataflow.

        Parameters
        ----------
        `dataflow` : `str | Dataflow`
            The Dataflow Id of the Dataflow to retrieve.

        Returns
        -------
        `Dataflow`
            The specified Dataflow.

        Notes
        -----
        The ExportDataflowAsAdmin endpoint returns a json file. For consistency
        with the rest of pbipy, this method returns a `Dataflow` object instead
        of the file itself. To access the json file, use the `Dataflow.raw`
        property.

        """

        if isinstance(dataflow, Dataflow):
            return Dataflow

        resource = self.base_path + f"/dataflows/{dataflow}/export"
        raw = _utils.get_raw(
            resource,
            self.session,
        )

        dataflow = Dataflow(
            raw.get("objectId"),
            self.session,
            group_id=None,
            raw=raw,
        )

        return dataflow

    def dataflows(
        self,
        group: str | Group = None,
        filter: str = None,
        skip: int = None,
        top: int = None,
    ) -> list[Dataflow]:
        """
        Returns a list of dataflows for the organization or specified Workspace
        (Group).

        Parameters
        ----------
        `group` : `str | Group`, optional
            Group Id or `Group` object to target. If not provided, then
            all Dataflows for the Organization will be returned.
        `filter` : `str`, optional
            Filters the results, based on a boolean condition.
        `skip` : `int`, optional
            Skips the first n results.
        `top` : `int`, optional
            Returns only the first n results.

        Returns
        -------
        `list[Dataflow]`
            List of Dataflows for the Organization or specified Workspace.

        """

        group_id = None

        if group:
            if isinstance(group, Group):
                group_id = group.id
            else:
                group_id = group

            path = f"/groups/{group_id}/dataflows"
        else:
            path = "/dataflows"

        params = {
            "$filter": filter,
            "$skip": skip,
            "$top": top,
        }

        url = self.base_path + path
        raw = _utils.get_raw(
            url,
            self.session,
            params=params,
        )

        dataflows = [
            Dataflow(
                id=dataflow_js.get("objectId"),
                session=self.session,
                group_id=dataflow_js.get("workspaceId", group_id),
                raw=dataflow_js,
            )
            for dataflow_js in raw
        ]

        return dataflows

    def dataflow_datasources(
        self,
        dataflow: str | Dataflow,
    ) -> list[dict]:
        """
        Return a list of Datasources for the specified Dataflow.

        Returns
        -------
        `list[dict]`
            The list of Datasources associated with the Dataflow.

        """

        path = _utils.build_path("/dataflows/{}/datasources", dataflow)
        url = self.base_path + path
        raw = _utils.get_raw(
            url,
            self.session,
        )

        return raw

    def dataflow_upstream_dataflows(
        self,
        dataflow: str | Dataflow,
        group: str | Group,
    ) -> list[dict]:
        """
        Returns a list of upstream dataflows for the specified Dataflow.

        Parameters
        ----------
        `dataflow` : `str | Dataflow`
            The Dataflow Id or `Dataflow` object to target.
        `group` : `str | Group`
            The Group Id or `Group` object where the target Dataflow resides.

        Returns
        -------
        `list[dict]`
            List of upstream dataflows.

        """

        path = _utils.build_path(
            "/groups/{}/dataflows/{}/upstreamDataflows", group, dataflow
        )
        url = self.base_path + path
        raw = _utils.get_raw(
            url,
            self.session,
        )

        return raw

    def dataflow_users(
        self,
        dataflow: str | Dataflow,
    ) -> list[dict]:
        """
        Returns a list of users that have access to the specified Dataflow.

        Parameters
        ----------
        `dataflow` : `str | Dataflow`
            The Dataflow Id or `Dataflow` object to retrieve the users for.

        Returns
        -------
        `list[dict]`
            List of users with access to the Dataflow.

        """

        path = _utils.build_path("/dataflows/{}/users", dataflow)
        url = self.base_path + path
        raw = _utils.get_raw(
            url,
            self.session,
        )

        return raw

    def datasets(
        self,
        group: str | Group = None,
        expand: str = None,
        filter: str = None,
        skip: int = None,
        top: int = None,
    ) -> list[Dataset]:
        """
        Returns a list of datasets for the Organization or specified Workspace
        (Group).

        Parameters
        ----------
        `group` : `str | Group`, optional
            Group Id or `Group` object to target. If provided then this
            method will return Datasets for the specified Group and not the
            Organization.
        `expand` : `str`, optional
            Expands related entities inline. If no `group` argument was
            provided, then this argument is ignored.
        `filter` : `str`, optional
            Filters the results, based on a boolean condition.
        `skip` : `int`, optional
            Skips the first n results.
        `top` : `int`, optional
            Returns only the first n results.

        Returns
        -------
        `list[Dataset]`
            List of Datasets for the Organization or specified Workspace.

        """

        params = {
            "$expand": expand,
            "$filter": filter,
            "$skip": skip,
            "$top": top,
        }

        # The non-group endpoint doesn't support $expand
        if group is None:
            params.pop("$expand")

        # Avoid referencing error in dataset_js.get("workspaceId", group_id)
        group_id = None

        if group:
            if isinstance(group, Group):
                group_id = group.id
            else:
                group_id = group

            path = f"/groups/{group_id}/datasets"
        else:
            path = "/datasets"

        url = self.base_path + path
        raw = _utils.get_raw(
            url,
            self.session,
            params=params,
        )

        datasets = [
            Dataset(
                id=dataset_js.get("id"),
                session=self.session,
                group_id=dataset_js.get("workspaceId", group_id),
                raw=dataset_js,
            )
            for dataset_js in raw
        ]

        return datasets

    def dataset_datasources(
        self,
        dataset: str | Dataset,
    ) -> list[dict]:
        """
        Returns a list of data sources for the specified dataset.

        Parameters
        ----------
        `dataset` : `str | Dataset`
            Dataset Id or `Dataset` object to target.

        Returns
        -------
        `list[dict]`
            List of Datasources for the specified Dataset.

        """

        path = _utils.build_path("/datasets/{}/datasources", dataset)
        url = self.base_path + path
        raw = _utils.get_raw(
            url,
            self.session,
        )

        return raw

    def dataset_users(
        self,
        dataset: str | Dataset,
    ) -> list[dict]:
        """
        Return a list of users that have access to the specified Dataset.

        Parameters
        ----------
        `dataset` : `str | Dataset`
            Dataset Id or `Dataset` object to target.

        Returns
        -------
        `list[dict]`
            List of Users with access to the specified Dataset.

        """

        path = _utils.build_path("/datasets/{}/users", dataset)
        url = self.base_path + path
        raw = _utils.get_raw(
            url,
            self.session,
        )

        return raw

    def datasets_upstream_dataflows(
        self,
        group: str | Group,
    ) -> list[dict]:
        """
        Return a list of upstream dataflows for datasets from the specified
        Workspace (Group).

        Parameters
        ----------
        `group` : `str | Group`
            Group Id or `Group` object to target.

        Returns
        -------
        `list[dict]`
            List of Dataset to Dataflow Links for the specified Workspace.
            Each element in this list represents a link between a Dataset
            and Dataflow.

        """

        path = _utils.build_path("/groups/{}/datasets/upstreamDataflows", group)
        url = self.base_path + path
        raw = _utils.get_raw(
            url,
            self.session,
        )

        return raw

    def group(
        self,
        group: str | Group,
        expand: str = None,
    ) -> Group:
        """
        Returns a workspace for the organization.

        Parameters
        ----------
        `group` : `str | Group`
            Group Id or `Group` object to target.
        `expand` : `str`, optional
            Accepts a comma-separated list of data types, which will be
            expanded inline in the response. Supports `users`, `reports`,
            `dashboards`, `datasets`, `dataflows`, and `workbooks`.

        Returns
        -------
        `Group`
            The specified Group for the Organization.

        """

        params = {"$expand": expand}

        path = _utils.build_path("/groups/{}", group)
        url = self.base_path + path

        raw = _utils.get_raw(
            url,
            self.session,
            params=params,
        )

        return Group(raw.get("id"), self.session, raw=raw)

    def groups(
        self,
        top: int = 5000,
        expand: str = None,
        filter: str = None,
        skip: int = None,
    ) -> list[Group]:
        """
        Returns a list of Workspaces (Groups) for the Organization.

        Parameters
        ----------
        `top` : `int`, optional
            Returns only the first n results. This parameter is considered
            mandatory by the end point and must be in the range of 1-5000.
            If not provided, will default to 5000, the maximum value permitted.
        `expand` : `str`, optional
            Accepts a comma-separated list of data types, which will be
            expanded inline in the response. Supports `users`, `reports`,
            `dashboards`, `datasets`, `dataflows`, and `workbooks`.
        `filter` : `str`, optional
            Filters the results based on a boolean condition, e.g.,
            `state eq 'Deleted'`
        `skip` : `int`, optional
            Skips the first n results. Use with top to fetch results beyond
            the first 5000.

        Returns
        -------
        `list[Group]`
            List of Workspaces (Groups) for the Organization.

        """

        params = {
            "$expand": expand,
            "$filter": filter,
            "$top": top,
            "$skip": skip,
        }

        url = self.base_path + "/groups"
        raw = _utils.get_raw(
            url,
            self.session,
            params=params,
        )

        groups = [
            Group(
                group_js.get("id"),
                session=self.session,
                raw=group_js,
            )
            for group_js in raw
        ]

        return groups

    def group_users(
        self,
        group: str | Group,
    ) -> list[dict]:
        """
        Returns a list of users that have access to the specified Workspace (Group).

        Parameters
        ----------
        `group` : `str | Group`
            Group Id or `Group` object to target.

        Returns
        -------
        `list[dict]`
            List of Group Users and associated details and permissions.

        """

        path = _utils.build_path("/groups/{}/users", group)
        url = self.base_path + path

        users = _utils.get_raw(
            url,
            self.session,
        )

        return users

    def add_group_user(
        self,
        group: str | Group,
        identifier: str,
        principal_type: str,
        group_user_access_right: str,
        display_name: str = None,
        email_address: str = None,
        graph_id: str = None,
        profile: dict = None,
        user_type: str = None,
    ) -> None:
        """
        Grants user permissions to the specified workspace. This API call
        only supports adding a user principal.

        Parameters
        ----------
        `group` : `str | Group`
            Group Id or `Group` object to add the user to.
        `identifier` : `str`
            Identifier of the principal.
        `principal_type` : `str`
            The principal type, e.g., "App", "Group", "None", or "User".
        `group_user_access_right` : `str`
            The access right (permission level) that a user has on the
            workspace, e.g. "Admin", "Contributor", "Member", "None",
            or "Viewer".
        `display_name` : `str`, optional
            Display name of the principal.
        `email_address` : `str`, optional
            Email address of the user.
        `graph_id` : `str`, optional
            Identifier of the principal in Microsoft Graph. Only available
            for admin APIs.
        `profile` : `dict`, optional
            A Power BI service principal profile. Only relevant for Power
            BI Embedded multi-tenancy solution.
        `user_type` : `str`, optional
            Type of the user.

        """

        group_user = {
            "identifier": identifier,
            "principalType": principal_type,
            "groupUserAccessRight": group_user_access_right,
            "displayName": display_name,
            "emailAddress": email_address,
            "graphId": graph_id,
            "profile": profile,
            "userType": user_type,
        }

        payload = _utils.remove_no_values(group_user)

        path = _utils.build_path("/groups/{}/users", group)
        url = self.base_path + path

        _utils.post(
            url,
            self.session,
            payload=payload,
        )

    def delete_group_user(
        self,
        group: str | Group,
        user: str,
        profile_id: str = None,
    ) -> None:
        """
        Removes user permissions from the specified workspace.

        Parameters
        ----------
        `group` : `str | Group`
            Group Id or `Group` object to target.
        `user` : `str`
            The user principal name (UPN) of the user to remove, e.g. `john@contoso.com`
        `profile_id` : `str`, optional
            The service principal profile ID to delete.

        """

        params = {"profileId": profile_id}

        path = _utils.build_path("/groups/{}/users/{}", group, user)
        url = self.base_path + path

        _utils.delete(
            url,
            self.session,
            params=params,
        )

    def restore_group(
        self,
        group: str | Group,
        email_address: str,
        name: str = None,
    ) -> None:
        """
        Restores a deleted Workspace (Group).

        Parameters
        ----------
        `group` : `str | Group`
            Group Id or `Group` object of the deleted Workspace.
        `email_address` : `str`
            The email address of the owner of the group to be restored.
        `name` : `str`, optional
            The name of the workspace to be restored.

        """

        request_body = {
            "emailAddress": email_address,
            "name": name,
        }
        request_body = _utils.remove_no_values(request_body)

        path = _utils.build_path("/groups/{}/restore", group)
        url = self.base_path + path

        _utils.post(
            url,
            self.session,
            payload=request_body,
        )

    def update_group(
        self,
        group: str | Group,
        name: str = None,
        description: str = None,
        **kwargs,
    ) -> None:
        """
        Updates the properties of the specified workspace.

        Only the name, description and Log Analytics Workspace can be updated.
        The name must be unique inside an Organization. To unassign a Log
        Analytics Workspace, explicitly set the value to `None`.

        Parameters
        ----------
        `group` : `str | Group`
            Group Id or `Group` object to target.
        `name` : `str`, optional
            The updated name.
        `description` : `str`, optional
            The updated description.
        `log_analytics_workspace` : `dict | None`, optional
            The Log Analytics Workspace to assign to the Group. Use `None`
            to unassign the Log Analytics Workspace.

        Raises
        ------
        `ValueError`
            If no values to update were provided.

        """

        request_body = {
            "name": name,
            "description": description,
        }
        request_body = _utils.remove_no_values(request_body)

        # To disable, the caller passes in None. Excluding implies logAnalyticsWorkspace
        # remains unchanged.
        if "log_analytics_workspace" in kwargs:
            request_body.update(
                {
                    "logAnalyticsWorkspace": kwargs["log_analytics_workspace"],
                }
            )

        if request_body in [None, {}]:
            raise ValueError(
                "No options were provided to update. Please specify an option to update."
            )

        path = _utils.build_path("/groups/{}", group)
        url = self.base_path + path

        _utils.patch(
            url,
            self.session,
            payload=request_body,
        )

    def reports(
        self,
        group: str | Group = None,
        filter: str = None,
        skip: int = None,
        top: int = None,
    ) -> list[Report]:
        """
        Returns a list of reports for the organization or the specified
        Workspace (Group).

        Parameters
        ----------
        `group` : `str | Group`, optional
            Group Id or `Group` object where the reports reside. If not
            supplied then returns all reports for the organization.
        `filter` : `str`, optional
            Filters the results, based on a boolean condition.
        `skip` : `int`, optional
            Skips the first n results.
        `top` : `int`, optional
            Returns only the first n results.

        Returns
        -------
        `list[Report]`
            List of reports for the organization or specified workspace

        """

        group_id = None
        if group:
            if isinstance(group, Group):
                group_id = group.id
            else:
                group_id = group

        if group:
            path = _utils.build_path("/groups/{}/reports", group)
        else:
            path = "/reports"

        url = self.base_path + path
        params = {
            "$filter": filter,
            "$skip": skip,
            "$top": top,
        }
        raw = _utils.get_raw(
            url,
            self.session,
            params=params,
        )

        reports = [
            Report(
                id=report_js.get("id"),
                session=self.session,
                group_id=report_js.get("workspaceId", group_id),
                raw=report_js,
            )
            for report_js in raw
        ]

        return reports

    def report_subscriptions(
        self,
        report: str | Report,
    ) -> list[dict]:
        """
        Returns a list of Report subscriptions along with subscriber details.
        This is a preview API call.

        Parameters
        ----------
        `report` : `str | Report`
            Report Id or `Report` object to target.

        Returns
        -------
        `list[dict]`
            List of report subscriptions and their details.

        """

        path = _utils.build_path("/reports/{}/subscriptions", report)
        url = self.base_path + path

        raw = _utils.get_raw(
            url,
            self.session,
        )

        return raw

    def report_users(
        self,
        report: str | Report,
    ) -> list[dict]:
        """
        Returns a list of users that have access to the specified report.

        Parameters
        ----------
        `report` : `str | Report`
            Report Id or `Report` object to target.

        Returns
        -------
        `list[dict]`
            List of Report Users and associated details.

        """

        path = _utils.build_path("/reports/{}/users", report)
        url = self.base_path + path

        raw = _utils.get_raw(
            url,
            self.session,
        )

        return raw

    def workspaces(
        self,
        exclude_inactive_workspaces: bool = None,
        exclude_personal_workspaces: bool = None,
        modified_since: datetime = None,
    ) -> list[dict]:
        """
        Gets a list of Workspace IDs in the Organization.

        Parameters
        ----------
        `exclude_inactive_workspaces` : `bool`, optional
            Whether to exclude inactive Workspaces.
        `exclude_personal_workspaces` : `bool`, optional
            Whether to exclude personal workspaces.
        `modified_since` : `datetime`, optional
            Only include Workspaces modified after this datetime. If not
            supplied then all Workspaces in the Organization are returned.
            The datetime provided is assumed to be UTC and will be supplied
            to the PowerBI API without any conversion to UTC.

            The datetime specified must be in the range of 30 minutes (to
            allow workspace changes to take effect) to 30 days prior to the
            current time.

        Returns
        -------
        `list[dict]`
            List of Workspaces with each Workspace represented as a dict
            and a single "ID" key.

        Notes
        -----
        Wraps the `/admin/workspaces/modified` endpoint which forms part
        of the 'scanner APIs'.

        See: https://learn.microsoft.com/en-us/fabric/governance/metadata-scanning-overview

        """

        url = self.base_path + "/workspaces/modified"

        # Comment here: http://disq.us/p/2dsf5lg, the datetime should follow
        # the 'O' or 'o' standard format string of the C# ToString method:
        # yyyy'-'MM'-'dd'T'HH':'mm':'ss'.'fffffffK
        # See: https://learn.microsoft.com/en-us/dotnet/standard/base-types/standard-date-and-time-format-strings#the-round-trip-o-o-format-specifier

        if modified_since:
            modified_since_formatted = modified_since.strftime("%Y-%m-%dT%H:%M:%S.%f0Z")
        else:
            modified_since_formatted = modified_since

        params = {
            "excludeInActiveWorkspaces": exclude_inactive_workspaces,
            "excludePersonalWorkspaces": exclude_personal_workspaces,
            "modifiedSince": modified_since_formatted,
        }

        raw = _utils.get_raw(
            url,
            self.session,
            params=params,
        )

        return raw

    def initiate_scan(
        self,
        workspaces: list[str] | str,
        dataset_expressions: bool = None,
        dataset_schema: bool = None,
        datasource_details: bool = None,
        get_artifact_users: bool = None,
        lineage: bool = None,
    ) -> dict:
        """
        Initiates a Metadata Scan for the requested workspace or list of
        workspaces.

        Parameters
        ----------
        `workspaces` : `list[str] | str`
            Workspace ID, or list of Workspace IDs to be scanned.
        `dataset_expressions` : `bool`, optional
            Whether to return dataset expressions (DAX and Mashup queries).
            If you set this parameter to `True`, you must fully enable metadata
            scanning in order for data to be returned.
        `dataset_schema` : `bool`, optional
            Whether to return dataset schema (tables, columns and measures).
            If you set this parameter to true, you must fully enable metadata
            scanning in order for data to be returned.
        `datasource_details` : `bool`, optional
            Whether to return data source details.
        `get_artifact_users` : `bool`, optional
            Whether to return user details for a Power BI item (such as
            a report or a dashboard).
        `lineage` : `bool`, optional
            Whether to return lineage info (upstream dataflows, tiles,
            data source IDs).

        Returns
        -------
        `dict`
            Dict representation of the Scan Request. Use the `"id"` key to
            access the scan request id. To check the status of the scan
            request, use the `scan_status` method.

        Notes
        -----
        Wraps the `/admin/workspaces/getInfo` endpoint which forms part
        of the 'scanner APIs'.

        See: https://learn.microsoft.com/en-us/fabric/governance/metadata-scanning-overview

        """

        url = self.base_path + "/workspaces/getInfo"

        if isinstance(workspaces, str):
            workspaces = [workspaces]

        request_body = {
            "workspaces": workspaces,
        }

        params = {
            "datasetExpressions": dataset_expressions,
            "datasetSchema": dataset_schema,
            "datasourceDetails": datasource_details,
            "getArtifactUsers": get_artifact_users,
            "lineage": lineage,
        }

        raw = _utils.post_raw(
            url,
            self.session,
            payload=request_body,
            params=params,
        )

        return raw

    def scan_status(
        self,
        scan_id: str,
    ) -> dict:
        """
        Gets the scan status for the specified scan id.

        Parameters
        ----------
        `scan_id` : `str`
            The scan id to get the status for. To initiate a scan use the
            `initate_scan` method, which will initiate a metadata scan
            on the Power BI instance and return a scan request that includes
            a scan id.

        Returns
        -------
        `dict`
            Dict representation of the scan request. Use the `"status"`
            key to identify the scan status.

        Notes
        -----
        Wraps the `/admin/workspaces/scanStatus` endpoint which forms part
        of the 'scanner APIs'.

        See: https://learn.microsoft.com/en-us/fabric/governance/metadata-scanning-overview

        """

        path = f"/workspaces/scanStatus/{scan_id}"
        url = self.base_path + path

        raw = _utils.get_raw(
            url,
            self.session,
        )

        return raw

    def scan_result(
        self,
        scan_id: str,
    ) -> dict:
        """
        Gets the scan result for the specified scan id. This should be
        called on a scan request with a `"Successful"` status. To determine
        the status of the scan, use the `scan_status` method with a scan
        id.

        Parameters
        ----------
        `scan_id` : `str`
            The scan id to get the scan result for.

        Returns
        -------
        `dict`
            The scan result (Workspace Info Details) as a dict.

        Notes
        -----
        Wraps the `/admin/workspaces/scanResult` endpoint which forms part
        of the 'scanner APIs'.

        See: https://learn.microsoft.com/en-us/fabric/governance/metadata-scanning-overview

        """

        path = f"/workspaces/scanResult/{scan_id}"
        url = self.base_path + path

        raw = _utils.get_raw(
            url,
            self.session,
        )

        return raw
